import os
import json
import webbrowser
import httplib2
import oauth2client.clientsecrets as clientsecrets
import threading

from googleapiclient.discovery import build
from functools import wraps
from oauth2client.service_account import ServiceAccountCredentials
from oauth2client.client import FlowExchangeError
from oauth2client.client import AccessTokenRefreshError
from oauth2client.client import OAuth2WebServerFlow
from oauth2client.client import OOB_CALLBACK_URN
from oauth2client.contrib.dictionary_storage import DictionaryStorage
from oauth2client.file import Storage
from oauth2client.tools import ClientRedirectHandler
from oauth2client.tools import ClientRedirectServer
from oauth2client._helpers import scopes_to_string
from .apiattr import ApiAttribute
from .apiattr import ApiAttributeMixin
from .settings import LoadSettingsFile
from .settings import ValidateSettings
from .settings import SettingsError
from .settings import InvalidConfigError


class AuthError(Exception):
    """Base error for authentication/authorization errors."""


class InvalidCredentialsError(IOError):
    """Error trying to read credentials file."""


class AuthenticationRejected(AuthError):
    """User rejected authentication."""


class AuthenticationError(AuthError):
    """General authentication error."""


class RefreshError(AuthError):
    """Access token refresh error."""


def LoadAuth(decoratee):
    """Decorator to check if the auth is valid and loads auth if not."""

    @wraps(decoratee)
    def _decorated(self, *args, **kwargs):
        # Initialize auth if needed.
        if self.auth is None:
            self.auth = GoogleAuth()
        # Re-create access token if it expired.
        if self.auth.access_token_expired:
            if getattr(self.auth, "auth_method", False) == "service":
                self.auth.ServiceAuth()
            else:
                self.auth.LocalWebserverAuth()

        # Initialise service if not built yet.
        if self.auth.service is None:
            self.auth.Authorize()

        # Ensure that a thread-safe HTTP object is provided.
        if (
            kwargs is not None
            and "param" in kwargs
            and kwargs["param"] is not None
            and "http" in kwargs["param"]
            and kwargs["param"]["http"] is not None
        ):
            self.http = kwargs["param"]["http"]
            del kwargs["param"]["http"]

        else:
            # If HTTP object not specified, create or resuse an HTTP
            # object from the thread local storage.
            if not getattr(self.auth.thread_local, "http", None):
                self.auth.thread_local.http = self.auth.Get_Http_Object()
            self.http = self.auth.thread_local.http

        return decoratee(self, *args, **kwargs)

    return _decorated


def CheckServiceAuth(decoratee):
    """Decorator to authorize service account."""

    @wraps(decoratee)
    def _decorated(self, *args, **kwargs):
        self.auth_method = "service"
        dirty = False
        save_credentials = self.settings.get("save_credentials")
        if self.credentials is None and save_credentials:
            self.LoadCredentials()
        if self.credentials is None:
            decoratee(self, *args, **kwargs)
            self.Authorize()
            dirty = True
        elif self.access_token_expired:
            self.Refresh()
            dirty = True
        self.credentials.set_store(self._default_storage)
        if dirty and save_credentials:
            self.SaveCredentials()

    return _decorated


def CheckAuth(decoratee):
    """Decorator to check if it requires OAuth2 flow request."""

    @wraps(decoratee)
    def _decorated(self, *args, **kwargs):
        dirty = False
        code = None
        save_credentials = self.settings.get("save_credentials")
        if self.credentials is None and save_credentials:
            self.LoadCredentials()
        if self.flow is None:
            self.GetFlow()
        if self.credentials is None:
            code = decoratee(self, *args, **kwargs)
            dirty = True
        else:
            if self.access_token_expired:
                if self.credentials.refresh_token is not None:
                    self.Refresh()
                else:
                    code = decoratee(self, *args, **kwargs)
                dirty = True
        if code is not None:
            self.Auth(code)
        self.credentials.set_store(self._default_storage)
        if dirty and save_credentials:
            self.SaveCredentials()

    return _decorated


class GoogleAuth(ApiAttributeMixin):
    """Wrapper class for oauth2client library in google-api-python-client.

    Loads all settings and credentials from one 'settings.yaml' file
    and performs common OAuth2.0 related functionality such as authentication
    and authorization.
    """

    DEFAULT_SETTINGS = {
        "client_config_backend": "file",
        "client_config_file": "client_secrets.json",
        "save_credentials": False,
        "oauth_scope": ["https://www.googleapis.com/auth/drive"],
    }
    CLIENT_CONFIGS_LIST = [
        "client_id",
        "client_secret",
        "auth_uri",
        "token_uri",
        "revoke_uri",
        "redirect_uri",
    ]
    SERVICE_CONFIGS_LIST = ["client_user_email"]
    settings = ApiAttribute("settings")
    client_config = ApiAttribute("client_config")
    flow = ApiAttribute("flow")
    credentials = ApiAttribute("credentials")
    http = ApiAttribute("http")
    service = ApiAttribute("service")
    auth_method = ApiAttribute("auth_method")

    def __init__(
        self, settings_file="settings.yaml", http_timeout=None, settings=None
    ):
        """Create an instance of GoogleAuth.

        :param settings_file: path of settings file. 'settings.yaml' by default.
        :type settings_file: str.
        :param settings: settings dict.
        :type settings: dict.
        """
        self.http_timeout = http_timeout
        ApiAttributeMixin.__init__(self)
        self.thread_local = threading.local()
        self.client_config = {}

        if settings is None and settings_file:
            try:
                settings = LoadSettingsFile(settings_file)
            except SettingsError:
                pass

        self.settings = settings or self.DEFAULT_SETTINGS
        ValidateSettings(self.settings)

        storages, default = self._InitializeStoragesFromSettings()
        self._storages = storages
        self._default_storage = default

    @property
    def access_token_expired(self):
        """Checks if access token doesn't exist or is expired.

        :returns: bool -- True if access token doesn't exist or is expired.
        """
        if self.credentials is None:
            return True
        return self.credentials.access_token_expired

    @CheckAuth
    def LocalWebserverAuth(
        self,
        host_name="localhost",
        port_numbers=None,
        launch_browser=True,
        bind_addr=None,
    ):
        """Authenticate and authorize from user by creating local web server and
        retrieving authentication code.

        This function is not for web server application. It creates local web
        server for user from standalone application.

        If GDRIVE_NON_INTERACTIVE environment variable is set, this function
        raises AuthenticationError.

        :param host_name: host name of the local web server.
        :type host_name: str.
        :param port_numbers: list of port numbers to be tried to used.
        :type port_numbers: list.
        :param launch_browser: should browser be launched automatically
        :type launch_browser: bool
        :param bind_addr: optional IP address for the local web server to listen on.
            If not specified, it will listen on the address specified in the
            host_name parameter.
        :type bind_addr: str.
        :returns: str -- code returned from local web server
        :raises: AuthenticationRejected, AuthenticationError
        """
        if os.getenv("GDRIVE_NON_INTERACTIVE"):
            raise AuthenticationError(
                "Non interactive mode (GDRIVE_NON_INTERACTIVE env) is enabled"
            )

        if port_numbers is None:
            port_numbers = [
                8080,
                8090,
            ]  # Mutable objects should not be default
            # values, as each call's changes are global.
        success = False
        port_number = 0
        for port in port_numbers:
            port_number = port
            try:
                httpd = ClientRedirectServer(
                    (bind_addr or host_name, port), ClientRedirectHandler
                )
            except OSError:
                pass
            else:
                success = True
                break
        if success:
            oauth_callback = f"http://{host_name}:{port_number}/"
        else:
            print(
                "Failed to start a local web server. Please check your firewall"
            )
            print(
                "settings and locally running programs that may be blocking or"
            )
            print("using configured ports. Default ports are 8080 and 8090.")
            raise AuthenticationError()
        self.flow.redirect_uri = oauth_callback
        authorize_url = self.GetAuthUrl()
        if launch_browser:
            webbrowser.open(authorize_url, new=1, autoraise=True)
            print("Your browser has been opened to visit:")
        else:
            print("Open your browser to visit:")
        print()
        print("    " + authorize_url)
        print()
        httpd.handle_request()
        if "error" in httpd.query_params:
            print("Authentication request was rejected")
            raise AuthenticationRejected("User rejected authentication")
        if "code" in httpd.query_params:
            return httpd.query_params["code"]
        else:
            print(
                'Failed to find "code" in the query parameters of the redirect.'
            )
            print("Try command-line authentication")
            raise AuthenticationError("No code found in redirect")

    @CheckAuth
    def CommandLineAuth(self):
        """Authenticate and authorize from user by printing authentication url
        retrieving authentication code from command-line.

        :returns: str -- code returned from commandline.
        """
        self.flow.redirect_uri = OOB_CALLBACK_URN
        authorize_url = self.GetAuthUrl()
        print("Go to the following link in your browser:")
        print()
        print("    " + authorize_url)
        print()
        return input("Enter verification code: ").strip()

    @CheckServiceAuth
    def ServiceAuth(self):
        """Authenticate and authorize using P12 private key, client id
        and client email for a Service account.
        :raises: AuthError, InvalidConfigError
        """
        if set(self.SERVICE_CONFIGS_LIST) - set(self.client_config):
            self.LoadServiceConfigSettings()
        scopes = scopes_to_string(self.settings["oauth_scope"])
        keyfile_name = self.client_config.get("client_json_file_path")
        keyfile_dict = self.client_config.get("client_json_dict")
        keyfile_json = self.client_config.get("client_json")

        if not keyfile_dict and keyfile_json:
            # Compensating for missing ServiceAccountCredentials.from_json_keyfile
            keyfile_dict = json.loads(keyfile_json)

        if keyfile_dict:
            self.credentials = (
                ServiceAccountCredentials.from_json_keyfile_dict(
                    keyfile_dict=keyfile_dict, scopes=scopes
                )
            )
        elif keyfile_name:
            self.credentials = (
                ServiceAccountCredentials.from_json_keyfile_name(
                    filename=keyfile_name, scopes=scopes
                )
            )
        else:
            service_email = self.client_config["client_service_email"]
            file_path = self.client_config["client_pkcs12_file_path"]
            self.credentials = ServiceAccountCredentials.from_p12_keyfile(
                service_account_email=service_email,
                filename=file_path,
                scopes=scopes,
            )

        user_email = self.client_config.get("client_user_email")
        if user_email:
            self.credentials = self.credentials.create_delegated(
                sub=user_email
            )

    def _InitializeStoragesFromSettings(self):
        result = {"file": None, "dictionary": None}
        backend = self.settings.get("save_credentials_backend")
        save_credentials = self.settings.get("save_credentials")
        if backend == "file":
            credentials_file = self.settings.get("save_credentials_file")
            if credentials_file is None:
                raise InvalidConfigError(
                    "Please specify credentials file to read"
                )
            result[backend] = Storage(credentials_file)
        elif backend == "dictionary":
            creds_dict = self.settings.get("save_credentials_dict")
            if creds_dict is None:
                raise InvalidConfigError("Please specify credentials dict")

            creds_key = self.settings.get("save_credentials_key")
            if creds_key is None:
                raise InvalidConfigError("Please specify credentials key")

            result[backend] = DictionaryStorage(creds_dict, creds_key)
        elif save_credentials:
            raise InvalidConfigError(
                "Unknown save_credentials_backend: %s" % backend
            )
        return result, result.get(backend)

    def LoadCredentials(self, backend=None):
        """Loads credentials or create empty credentials if it doesn't exist.

        :param backend: target backend to save credential to.
        :type backend: str.
        :raises: InvalidConfigError
        """
        if backend is None:
            backend = self.settings.get("save_credentials_backend")
            if backend is None:
                raise InvalidConfigError("Please specify credential backend")
        if backend == "file":
            self.LoadCredentialsFile()
        elif backend == "dictionary":
            self._LoadCredentialsDictionary()
        else:
            raise InvalidConfigError("Unknown save_credentials_backend")

    def LoadCredentialsFile(self, credentials_file=None):
        """Loads credentials or create empty credentials if it doesn't exist.

        Loads credentials file from path in settings if not specified.

        :param credentials_file: path of credentials file to read.
        :type credentials_file: str.
        :raises: InvalidConfigError, InvalidCredentialsError
        """
        if credentials_file is None:
            self._default_storage = self._storages["file"]
            if self._default_storage is None:
                raise InvalidConfigError(
                    "Backend `file` is not configured, specify "
                    "credentials file to read in the settings "
                    "file or pass an explicit value"
                )
        else:
            self._default_storage = Storage(credentials_file)

        try:
            self.credentials = self._default_storage.get()
        except OSError:
            raise InvalidCredentialsError(
                "Credentials file cannot be symbolic link"
            )

        if self.credentials:
            self.credentials.set_store(self._default_storage)

    def _LoadCredentialsDictionary(self):
        self._default_storage = self._storages["dictionary"]
        if self._default_storage is None:
            raise InvalidConfigError(
                "Backend `dictionary` is not configured, specify "
                "credentials dict and key to read in the settings file"
            )

        self.credentials = self._default_storage.get()

        if self.credentials:
            self.credentials.set_store(self._default_storage)

    def SaveCredentials(self, backend=None):
        """Saves credentials according to specified backend.

        If you have any specific credentials backend in mind, don't use this
        function and use the corresponding function you want.

        :param backend: backend to save credentials.
        :type backend: str.
        :raises: InvalidConfigError
        """
        if backend is None:
            backend = self.settings.get("save_credentials_backend")
            if backend is None:
                raise InvalidConfigError("Please specify credential backend")
        if backend == "file":
            self.SaveCredentialsFile()
        elif backend == "dictionary":
            self._SaveCredentialsDictionary()
        else:
            raise InvalidConfigError("Unknown save_credentials_backend")

    def SaveCredentialsFile(self, credentials_file=None):
        """Saves credentials to the file in JSON format.

        :param credentials_file: destination to save file to.
        :type credentials_file: str.
        :raises: InvalidConfigError, InvalidCredentialsError
        """
        if self.credentials is None:
            raise InvalidCredentialsError("No credentials to save")

        if credentials_file is None:
            storage = self._storages["file"]
            if storage is None:
                raise InvalidConfigError(
                    "Backend `file` is not configured, specify "
                    "credentials file to read in the settings "
                    "file or pass an explicit value"
                )
        else:
            storage = Storage(credentials_file)

        try:
            storage.put(self.credentials)
        except OSError:
            raise InvalidCredentialsError(
                "Credentials file cannot be symbolic link"
            )

    def _SaveCredentialsDictionary(self):
        if self.credentials is None:
            raise InvalidCredentialsError("No credentials to save")

        storage = self._storages["dictionary"]
        if storage is None:
            raise InvalidConfigError(
                "Backend `dictionary` is not configured, specify "
                "credentials dict and key to write in the settings file"
            )

        storage.put(self.credentials)

    def LoadClientConfig(self, backend=None):
        """Loads client configuration according to specified backend.

        If you have any specific backend to load client configuration from in mind,
        don't use this function and use the corresponding function you want.

        :param backend: backend to load client configuration from.
        :type backend: str.
        :raises: InvalidConfigError
        """
        if backend is None:
            backend = self.settings.get("client_config_backend")
            if backend is None:
                raise InvalidConfigError(
                    "Please specify client config backend"
                )
        if backend == "file":
            self.LoadClientConfigFile()
        elif backend == "settings":
            self.LoadClientConfigSettings()
        elif backend == "service":
            self.LoadServiceConfigSettings()
        else:
            raise InvalidConfigError("Unknown client_config_backend")

    def LoadClientConfigFile(self, client_config_file=None):
        """Loads client configuration file downloaded from APIs console.

        Loads client config file from path in settings if not specified.

        :param client_config_file: path of client config file to read.
        :type client_config_file: str.
        :raises: InvalidConfigError
        """
        if client_config_file is None:
            client_config_file = self.settings["client_config_file"]
        try:
            client_type, client_info = clientsecrets.loadfile(
                client_config_file
            )
        except clientsecrets.InvalidClientSecretsError as error:
            raise InvalidConfigError("Invalid client secrets file %s" % error)
        if client_type not in (
            clientsecrets.TYPE_WEB,
            clientsecrets.TYPE_INSTALLED,
        ):
            raise InvalidConfigError(
                "Unknown client_type of client config file"
            )

        # General settings.
        try:
            config_index = [
                "client_id",
                "client_secret",
                "auth_uri",
                "token_uri",
            ]
            for config in config_index:
                self.client_config[config] = client_info[config]

            self.client_config["revoke_uri"] = client_info.get("revoke_uri")
            self.client_config["redirect_uri"] = client_info["redirect_uris"][
                0
            ]
        except KeyError:
            raise InvalidConfigError("Insufficient client config in file")

        # Service auth related fields.
        service_auth_config = ["client_email"]
        try:
            for config in service_auth_config:
                self.client_config[config] = client_info[config]
        except KeyError:
            pass  # The service auth fields are not present, handling code can go here.

    def LoadServiceConfigSettings(self):
        """Loads client configuration from settings.
        :raises: InvalidConfigError
        """
        configs = [
            "client_json_file_path",
            "client_json_dict",
            "client_json",
            "client_pkcs12_file_path",
        ]

        for config in configs:
            value = self.settings["service_config"].get(config)
            if value:
                self.client_config[config] = value
                break
        else:
            raise InvalidConfigError(
                f"One of {configs} is required for service authentication"
            )

        if config == "client_pkcs12_file_path":
            self.SERVICE_CONFIGS_LIST.append("client_service_email")

        for config in self.SERVICE_CONFIGS_LIST:
            try:
                self.client_config[config] = self.settings["service_config"][
                    config
                ]
            except KeyError:
                err = "Insufficient service config in settings"
                err += f"\n\nMissing: {config} key."
                raise InvalidConfigError(err)

    def LoadClientConfigSettings(self):
        """Loads client configuration from settings file.

        :raises: InvalidConfigError
        """
        for config in self.CLIENT_CONFIGS_LIST:
            try:
                self.client_config[config] = self.settings["client_config"][
                    config
                ]
            except KeyError:
                raise InvalidConfigError(
                    "Insufficient client config in settings"
                )

    def GetFlow(self):
        """Gets Flow object from client configuration.

        :raises: InvalidConfigError
        """
        if not all(
            config in self.client_config for config in self.CLIENT_CONFIGS_LIST
        ):
            self.LoadClientConfig()
        constructor_kwargs = {
            "redirect_uri": self.client_config["redirect_uri"],
            "auth_uri": self.client_config["auth_uri"],
            "token_uri": self.client_config["token_uri"],
            "access_type": "online",
        }
        if self.client_config["revoke_uri"] is not None:
            constructor_kwargs["revoke_uri"] = self.client_config["revoke_uri"]
        self.flow = OAuth2WebServerFlow(
            self.client_config["client_id"],
            self.client_config["client_secret"],
            scopes_to_string(self.settings["oauth_scope"]),
            **constructor_kwargs,
        )
        if self.settings.get("get_refresh_token"):
            self.flow.params.update(
                {"access_type": "offline", "approval_prompt": "force"}
            )

    def Refresh(self):
        """Refreshes the access_token.

        :raises: RefreshError
        """
        if self.credentials is None:
            raise RefreshError("No credential to refresh.")
        if (
            self.credentials.refresh_token is None
            and self.auth_method != "service"
        ):
            raise RefreshError(
                "No refresh_token found."
                "Please set access_type of OAuth to offline."
            )
        if self.http is None:
            self.http = self._build_http()
        try:
            self.credentials.refresh(self.http)
        except AccessTokenRefreshError as error:
            raise RefreshError("Access token refresh failed: %s" % error)

    def GetAuthUrl(self):
        """Creates authentication url where user visits to grant access.

        :returns: str -- Authentication url.
        """
        if self.flow is None:
            self.GetFlow()
        return self.flow.step1_get_authorize_url()

    def Auth(self, code):
        """Authenticate, authorize, and build service.

        :param code: Code for authentication.
        :type code: str.
        :raises: AuthenticationError
        """
        self.Authenticate(code)
        self.Authorize()

    def Authenticate(self, code):
        """Authenticates given authentication code back from user.

        :param code: Code for authentication.
        :type code: str.
        :raises: AuthenticationError
        """
        if self.flow is None:
            self.GetFlow()
        try:
            self.credentials = self.flow.step2_exchange(code)
        except FlowExchangeError as e:
            raise AuthenticationError("OAuth2 code exchange failed: %s" % e)
        print("Authentication successful.")

    def _build_http(self):
        http = httplib2.Http(timeout=self.http_timeout)
        # 308's are used by several Google APIs (Drive, YouTube)
        # for Resumable Uploads rather than Permanent Redirects.
        # This asks httplib2 to exclude 308s from the status codes
        # it treats as redirects
        # See also: https://stackoverflow.com/a/59850170/298182
        try:
            http.redirect_codes = http.redirect_codes - {308}
        except AttributeError:
            # http.redirect_codes does not exist in previous versions
            # of httplib2, so pass
            pass
        return http

    def Authorize(self):
        """Authorizes and builds service.

        :raises: AuthenticationError
        """
        if self.access_token_expired:
            raise AuthenticationError(
                "No valid credentials provided to authorize"
            )

        if self.http is None:
            self.http = self._build_http()
        self.http = self.credentials.authorize(self.http)
        self.service = build(
            "drive", "v2", http=self.http, cache_discovery=False
        )

    def Get_Http_Object(self):
        """Create and authorize an httplib2.Http object. Necessary for
        thread-safety.
        :return: The http object to be used in each call.
        :rtype: httplib2.Http
        """
        http = self._build_http()
        http = self.credentials.authorize(http)
        return http



from io import StringIO
import os
import posixpath
import secrets
import uuid
from concurrent import futures

import pytest
import fsspec
from pydrive2.auth import GoogleAuth
from pydrive2.fs import GDriveFileSystem
from pydrive2.test.test_util import settings_file_path, setup_credentials
from pydrive2.test.test_util import GDRIVE_USER_CREDENTIALS_DATA

TEST_GDRIVE_REPO_BUCKET = "root"


@pytest.fixture(scope="module")
def base_remote_dir():
    path = TEST_GDRIVE_REPO_BUCKET + "/" + str(uuid.uuid4())
    return path


@pytest.fixture
def remote_dir(base_remote_dir):
    return base_remote_dir + "/" + str(uuid.uuid4())


@pytest.fixture(scope="module")
def service_auth(tmp_path_factory):
    setup_credentials()
    tmpdir = tmp_path_factory.mktemp("settings")
    auth = GoogleAuth(settings_file_path("default.yaml", wkdir=tmpdir))
    auth.ServiceAuth()
    return auth


@pytest.fixture(scope="module")
def fs_factory(base_remote_dir, service_auth):
    base_item = None
    GDriveFileSystem.cachable = False

    def _create_fs():
        nonlocal base_item
        _, base = base_remote_dir.split("/", 1)
        fs = GDriveFileSystem(base_remote_dir, service_auth)
        if base_item is None:
            base_item = fs._gdrive_create_dir("root", base)

        return fs, base_item

    yield _create_fs

    GDriveFileSystem.cachable = True
    fs = GDriveFileSystem(base_remote_dir, service_auth)
    fs.rm_file(base_remote_dir)


@pytest.fixture
def fs(fs_factory):
    return fs_factory()[0]


@pytest.mark.manual
def test_fs_oauth(base_remote_dir):
    GDriveFileSystem(
        base_remote_dir,
        client_id="47794215776-cd9ssb6a4vv5otkq6n0iadpgc4efgjb1.apps.googleusercontent.com",  # noqa: E501
        client_secret="i2gerGA7uBjZbR08HqSOSt9Z",
    )


def test_fs_service_json_file(base_remote_dir):
    creds = "credentials/fs.dat"
    setup_credentials(creds)
    GDriveFileSystem(
        base_remote_dir,
        use_service_account=True,
        client_json_file_path=creds,
    )


def test_fs_service_json(base_remote_dir):
    creds = os.environ[GDRIVE_USER_CREDENTIALS_DATA]
    GDriveFileSystem(
        base_remote_dir,
        use_service_account=True,
        client_json=creds,
    )


def test_info(fs, remote_dir):
    fs.touch(remote_dir + "/info/a.txt")
    fs.touch(remote_dir + "/info/b.txt")
    details = fs.info(remote_dir + "/info/a.txt")
    assert details["type"] == "file"
    assert details["name"] == remote_dir + "/info/a.txt"
    assert details["size"] == 0
    assert (
        details["checksum"] == fs.info(remote_dir + "/info/b.txt")["checksum"]
    )

    details = fs.info(remote_dir + "/info")
    assert details["type"] == "directory"
    assert details["name"] == remote_dir + "/info/"
    assert "checksum" not in details

    details = fs.info(remote_dir + "/info/")
    assert details["type"] == "directory"
    assert details["name"] == remote_dir + "/info/"


def test_move(fs, remote_dir):
    fs.touch(remote_dir + "/a.txt")
    initial_info = fs.info(remote_dir + "/a.txt")

    fs.move(remote_dir + "/a.txt", remote_dir + "/b.txt")
    secondary_info = fs.info(remote_dir + "/b.txt")

    assert not fs.exists(remote_dir + "/a.txt")
    assert fs.exists(remote_dir + "/b.txt")

    initial_info.pop("name")
    secondary_info.pop("name")
    assert initial_info == secondary_info


def test_rm(fs, remote_dir):
    fs.touch(remote_dir + "/a.txt")
    fs.rm(remote_dir + "/a.txt")
    assert not fs.exists(remote_dir + "/a.txt")

    fs.mkdir(remote_dir + "/dir")
    fs.touch(remote_dir + "/dir/a")
    fs.touch(remote_dir + "/dir/b")
    fs.mkdir(remote_dir + "/dir/c/")
    fs.touch(remote_dir + "/dir/c/a")
    fs.rm(remote_dir + "/dir", recursive=True)
    assert not fs.exists(remote_dir + "/dir/c/a")


def test_ls(fs, remote_dir):
    _, base = fs.split_path(remote_dir + "/dir/")
    fs._path_to_item_ids(base, create=True)
    assert fs.ls(remote_dir + "/dir/") == []

    files = set()
    for no in range(8):
        file = remote_dir + f"/dir/test_{no}"
        fs.touch(file)
        files.add(file)

    assert set(fs.ls(remote_dir + "/dir/")) == files

    dirs = fs.ls(remote_dir + "/dir/", detail=True)
    expected = [fs.info(file) for file in files]

    def by_name(details):
        return details["name"]

    dirs.sort(key=by_name)
    expected.sort(key=by_name)

    assert dirs == expected


def test_basic_ops_caching(fs_factory, remote_dir, mocker):
    # Internally we have to derefence names into IDs to call GDrive APIs
    # we are trying hard to cache those and make sure that operations like
    # exists, ls, find, etc. don't hit the API more than once per path

    # ListFile (_gdrive_list) is the main operation that we use to retieve file
    # metadata in all operations like find/ls/exist - etc. It should be fine as
    # a basic benchmark to count those.
    # Note: we can't count direct API calls since we have retries, also can't
    # count even direct calls to the GDrive client - for the same reason
    fs, _ = fs_factory()
    spy = mocker.spy(fs, "_gdrive_list")

    dir_path = remote_dir + "/a/b/c/"
    file_path = dir_path + "test.txt"
    fs.touch(file_path)

    assert spy.call_count == 5
    spy.reset_mock()

    fs.exists(file_path)
    assert spy.call_count == 1
    spy.reset_mock()

    fs.ls(remote_dir)
    assert spy.call_count == 1
    spy.reset_mock()

    fs.ls(dir_path)
    assert spy.call_count == 1
    spy.reset_mock()

    fs.find(dir_path)
    assert spy.call_count == 1
    spy.reset_mock()

    fs.find(remote_dir)
    assert spy.call_count == 1
    spy.reset_mock()


def test_ops_work_with_duplicate_names(fs_factory, remote_dir):
    fs, base_item = fs_factory()

    remote_dir_item = fs._gdrive_create_dir(
        base_item["id"], remote_dir.split("/")[-1]
    )
    dir_name = str(uuid.uuid4())
    dir1 = fs._gdrive_create_dir(remote_dir_item["id"], dir_name)
    dir2 = fs._gdrive_create_dir(remote_dir_item["id"], dir_name)

    # Two directories were created with the same name
    assert dir1["id"] != dir2["id"]

    dir_path = remote_dir + "/" + dir_name + "/"

    # ls returns both of them, even though the names are the same
    test_fs = fs
    result = test_fs.ls(remote_dir)
    assert len(result) == 2
    assert set(result) == {dir_path}

    # ls returns both of them, even though the names are the same
    test_fs, _ = fs_factory()
    result = test_fs.ls(remote_dir)
    assert len(result) == 2
    assert set(result) == {dir_path}

    for test_fs in [fs, fs_factory()[0]]:
        # find by default doesn't return dirs at all
        result = test_fs.find(remote_dir)
        assert len(result) == 0

    fs._gdrive_upload_fobj("a.txt", dir1["id"], StringIO(""))
    fs._gdrive_upload_fobj("b.txt", dir2["id"], StringIO(""))

    for test_fs in [fs, fs_factory()[0]]:
        # now we should have both files
        result = test_fs.find(remote_dir)
        assert len(result) == 2
        assert set(result) == {dir_path + file for file in ["a.txt", "b.txt"]}


def test_ls_non_existing_dir(fs, remote_dir):
    with pytest.raises(FileNotFoundError):
        fs.ls(remote_dir + "dir/")


def test_find(fs, fs_factory, remote_dir):
    fs.mkdir(remote_dir + "/dir")

    files = [
        "a",
        "b",
        "c/a",
        "c/b",
        "c/d/a",
        "c/d/b",
        "c/d/c",
        "c/d/f/a",
        "c/d/f/b",
    ]
    files = [remote_dir + "/dir/" + file for file in files]
    dirnames = {posixpath.dirname(file) for file in files}

    for dirname in dirnames:
        fs.mkdir(dirname)

    for file in files:
        fs.touch(file)

    for test_fs in [fs, fs_factory()[0]]:
        # Test for https://github.com/iterative/PyDrive2/issues/229
        # It must go first, so that we test with a cache miss as well
        assert set(test_fs.find(remote_dir + "/dir/c/d/")) == set(
            [
                file
                for file in files
                if file.startswith(remote_dir + "/dir/c/d/")
            ]
        )

        # General find test
        assert set(test_fs.find(remote_dir)) == set(files)

        find_results = test_fs.find(remote_dir, detail=True)
        info_results = [test_fs.info(file) for file in files]
        info_results = {content["name"]: content for content in info_results}
        assert find_results == info_results


def test_exceptions(fs, tmpdir, remote_dir):
    with pytest.raises(FileNotFoundError):
        with fs.open(remote_dir + "/a.txt"):
            ...

    with pytest.raises(FileNotFoundError):
        fs.copy(remote_dir + "/u.txt", remote_dir + "/y.txt")

    with pytest.raises(FileNotFoundError):
        fs.get_file(remote_dir + "/c.txt", tmpdir / "c.txt")


def test_open_rw(fs, remote_dir):
    data = b"dvc.org"

    with fs.open(remote_dir + "/a.txt", "wb") as stream:
        stream.write(data)

    with fs.open(remote_dir + "/a.txt") as stream:
        assert stream.read() == data


def test_concurrent_operations(fs, fs_factory, remote_dir):
    # Include an extra dir name to force upload operations creating it
    # this way we can also test that only a single directory is created
    # enven if multiple threads are uploading files into the same dir
    dir_name = secrets.token_hex(16)

    def create_random_file():
        name = secrets.token_hex(16)
        with fs.open(remote_dir + f"/{dir_name}/" + name, "w") as stream:
            stream.write(name)
        return name

    def read_random_file(name):
        with fs.open(remote_dir + f"/{dir_name}/" + name, "r") as stream:
            return stream.read()

    with futures.ThreadPoolExecutor() as executor:
        write_futures, _ = futures.wait(
            [executor.submit(create_random_file) for _ in range(64)],
            return_when=futures.ALL_COMPLETED,
        )
        write_names = {future.result() for future in write_futures}

        read_futures, _ = futures.wait(
            [executor.submit(read_random_file, name) for name in write_names],
            return_when=futures.ALL_COMPLETED,
        )
        read_names = {future.result() for future in read_futures}

        assert write_names == read_names

    # Test that only a single dir is cretead
    for test_fs in [fs, fs_factory()[0]]:
        results = test_fs.ls(remote_dir)
        assert results == [remote_dir + f"/{dir_name}/"]


def test_put_file(fs, tmpdir, remote_dir):
    src_file = tmpdir / "a.txt"
    with open(src_file, "wb") as file:
        file.write(b"data")

    fs.put_file(src_file, remote_dir + "/a.txt")

    with fs.open(remote_dir + "/a.txt") as stream:
        assert stream.read() == b"data"


def test_get_file(fs, tmpdir, remote_dir):
    src_file = tmpdir / "a.txt"
    dest_file = tmpdir / "b.txt"

    with open(src_file, "wb") as file:
        file.write(b"data")

    fs.put_file(src_file, remote_dir + "/a.txt")
    fs.get_file(remote_dir + "/a.txt", dest_file)
    assert dest_file.read() == "data"


def test_get_file_callback(fs, tmpdir, remote_dir):
    src_file = tmpdir / "a.txt"
    dest_file = tmpdir / "b.txt"

    with open(src_file, "wb") as file:
        file.write(b"data" * 10)

    fs.put_file(src_file, remote_dir + "/a.txt")
    callback = fsspec.Callback()
    fs.get_file(
        remote_dir + "/a.txt", dest_file, callback=callback, block_size=10
    )
    assert dest_file.read() == "data" * 10

    assert callback.size == 40
    assert callback.value == 40



import subprocess

from PyInstaller import __main__ as pyi_main


# NOTE: importlib.resources.contents is available in py3.7+, but due to how
# pyinstaller handles importlib, we need to use the importlib_resources
# backport if there are any resources methods that are not available in a given
# python version, which ends up being py<3.10
_APP_SOURCE = """
import sys
if sys.version_info >= (3, 10):
    from importlib.resources import contents
else:
    from importlib_resources import contents

import pydrive2.files


cache_files = contents(
    "googleapiclient.discovery_cache.documents"
)
assert len(cache_files) > 0
"""


def test_pyi_hook_google_api_client(tmp_path):
    app_name = "userapp"
    workpath = tmp_path / "build"
    distpath = tmp_path / "dist"
    app = tmp_path / f"{app_name}.py"
    app.write_text(_APP_SOURCE)
    pyi_main.run(
        [
            "--workpath",
            str(workpath),
            "--distpath",
            str(distpath),
            "--specpath",
            str(tmp_path),
            str(app),
        ],
    )
    subprocess.run([str(distpath / app_name / app_name)], check=True)



class ApiAttribute:
    """A data descriptor that sets and returns values."""

    def __init__(self, name):
        """Create an instance of ApiAttribute.

        :param name: name of this attribute.
        :type name: str.
        """
        self.name = name

    def __get__(self, obj, type=None):
        """Accesses value of this attribute."""
        return obj.attr.get(self.name)

    def __set__(self, obj, value):
        """Write value of this attribute."""
        obj.attr[self.name] = value
        if obj.dirty.get(self.name) is not None:
            obj.dirty[self.name] = True

    def __del__(self, obj=None):
        """Delete value of this attribute."""
        if not obj:
            return

        del obj.attr[self.name]
        if obj.dirty.get(self.name) is not None:
            del obj.dirty[self.name]


class ApiAttributeMixin:
    """Mixin to initialize required global variables to use ApiAttribute."""

    def __init__(self):
        self.attr = {}
        self.dirty = {}
        self.http = None  # Any element may make requests and will require this
        # field.


class ApiResource(dict):
    """Super class of all api resources.

    Inherits and behaves as a python dictionary to handle api resources.
    Save clean copy of metadata in self.metadata as a dictionary.
    Provides changed metadata elements to efficiently update api resources.
    """

    auth = ApiAttribute("auth")

    def __init__(self, *args, **kwargs):
        """Create an instance of ApiResource."""
        super().__init__()
        self.update(*args, **kwargs)
        self.metadata = dict(self)

    def __getitem__(self, key):
        """Overwritten method of dictionary.

        :param key: key of the query.
        :type key: str.
        :returns: value of the query.
        """
        return dict.__getitem__(self, key)

    def __setitem__(self, key, val):
        """Overwritten method of dictionary.

        :param key: key of the query.
        :type key: str.
        :param val: value of the query.
        """
        dict.__setitem__(self, key, val)

    def __repr__(self):
        """Overwritten method of dictionary."""
        dict_representation = dict.__repr__(self)
        return f"{type(self).__name__}({dict_representation})"

    def update(self, *args, **kwargs):
        """Overwritten method of dictionary."""
        BAD_URL_PREFIX = "https://www.googleapis.comhttps:"
        for k, v in dict(*args, **kwargs).items():
            if k == "downloadUrl" and v.startswith(BAD_URL_PREFIX):
                v = v.replace(BAD_URL_PREFIX, "https://www.googleapis.com", 1)
            self[k] = v

    def UpdateMetadata(self, metadata=None):
        """Update metadata and mark all of them to be clean."""
        if metadata:
            self.update(metadata)
        self.metadata = dict(self)

    def GetChanges(self):
        """Returns changed metadata elements to update api resources efficiently.

        :returns: dict -- changed metadata elements.
        """
        dirty = {}
        for key in self:
            if self.metadata.get(key) is None:
                dirty[key] = self[key]
            elif self.metadata[key] != self[key]:
                dirty[key] = self[key]
        return dirty


class ApiResourceList(ApiAttributeMixin, ApiResource):
    """Abstract class of all api list resources.

    Inherits ApiResource and builds iterator to list any API resource.
    """

    metadata = ApiAttribute("metadata")

    def __init__(self, auth=None, metadata=None):
        """Create an instance of ApiResourceList.

        :param auth: authorized GoogleAuth instance.
        :type auth: GoogleAuth.
        :param metadata: parameter to send to list command.
        :type metadata: dict.
        """
        ApiAttributeMixin.__init__(self)
        ApiResource.__init__(self)
        self.auth = auth
        self.UpdateMetadata()
        if metadata:
            self.update(metadata)

    def __iter__(self):
        """Returns iterator object.

        :returns: ApiResourceList -- self
        """
        return self

    def __next__(self):
        """Make API call to list resources and return them.

        Auto updates 'pageToken' every time it makes API call and
        raises StopIteration when it reached the end of iteration.

        :returns: list -- list of API resources.
        :raises: StopIteration
        """
        if "pageToken" in self and self["pageToken"] is None:
            raise StopIteration
        result = self._GetList()
        self["pageToken"] = self.metadata.get("nextPageToken")
        return result

    def GetList(self):
        """Get list of API resources.

        If 'maxResults' is not specified, it will automatically iterate through
        every resources available. Otherwise, it will make API call once and
        update 'pageToken'.

        :returns: list -- list of API resources.
        """
        if self.get("maxResults") is None:
            self["maxResults"] = 1000
            result = []
            for x in self:
                result.extend(x)
            del self["maxResults"]
            return result
        else:
            return next(self)

    def _GetList(self):
        """Helper function which actually makes API call.

        Should be overwritten.

        :raises: NotImplementedError
        """
        raise NotImplementedError

    def Reset(self):
        """Resets current iteration"""
        if "pageToken" in self:
            del self["pageToken"]



import unittest

from pydrive2.auth import GoogleAuth
from pydrive2.drive import GoogleDrive
from pydrive2.test.test_util import (
    pydrive_retry,
    setup_credentials,
    settings_file_path,
)


class ApiAttributeTest(unittest.TestCase):
    """Test ApiAttr functions."""

    @classmethod
    def setup_class(cls):
        setup_credentials()

    def test_UpdateMetadataNotInfinitelyNesting(self):
        # Verify 'metadata' field present.
        self.assertTrue(self.file1.metadata is not None)
        pydrive_retry(self.file1.UpdateMetadata)

        # Verify 'metadata' field still present.
        self.assertTrue(self.file1.metadata is not None)
        # Ensure no 'metadata' field in 'metadata' (i.e. nested).
        self.assertTrue("metadata" not in self.file1.metadata)

    def setUp(self):
        ga = GoogleAuth(settings_file_path("default.yaml"))
        ga.ServiceAuth()
        self.drive = GoogleDrive(ga)
        self.file1 = self.drive.CreateFile()
        pydrive_retry(self.file1.Upload)

    def tearDown(self):
        pydrive_retry(self.file1.Delete)


if __name__ == "__main__":
    unittest.main()



import io
import os
import mimetypes
import json

from googleapiclient import errors
from googleapiclient.http import MediaIoBaseUpload
from googleapiclient.http import MediaIoBaseDownload
from googleapiclient.http import DEFAULT_CHUNK_SIZE
from functools import wraps

from .apiattr import ApiAttribute
from .apiattr import ApiAttributeMixin
from .apiattr import ApiResource
from .apiattr import ApiResourceList
from .auth import LoadAuth

BLOCK_SIZE = 1024
# Usage: MIME_TYPE_TO_BOM['<Google Drive mime type>']['<download mimetype>'].
MIME_TYPE_TO_BOM = {
    "application/vnd.google-apps.document": {"text/plain": "\ufeff".encode()}
}


class FileNotUploadedError(RuntimeError):
    """Error trying to access metadata of file that is not uploaded."""


class ApiRequestError(IOError):
    def __init__(self, http_error):
        assert isinstance(http_error, errors.HttpError)
        content = json.loads(http_error.content.decode("utf-8"))
        self.error = content.get("error", {}) if content else {}

        # Initialize args for backward compatibility
        super().__init__(http_error)

    def GetField(self, field):
        """Returns the `field` from the first error"""
        return self.error.get("errors", [{}])[0].get(field, "")


class FileNotDownloadableError(RuntimeError):
    """Error trying to download file that is not downloadable."""


def LoadMetadata(decoratee):
    """Decorator to check if the file has metadata and fetches it if not.

    :raises: ApiRequestError, FileNotUploadedError
    """

    @wraps(decoratee)
    def _decorated(self, *args, **kwargs):
        if not self.uploaded:
            self.FetchMetadata()
        return decoratee(self, *args, **kwargs)

    return _decorated


class GoogleDriveFileList(ApiResourceList):
    """Google Drive FileList instance.

    Equivalent to Files.list() in Drive APIs.
    """

    def __init__(self, auth=None, param=None):
        """Create an instance of GoogleDriveFileList."""
        super().__init__(auth=auth, metadata=param)

    @LoadAuth
    def _GetList(self):
        """Overwritten method which actually makes API call to list files.

        :returns: list -- list of pydrive2.files.GoogleDriveFile.
        """
        # Teamdrive support
        self["supportsAllDrives"] = True
        self["includeItemsFromAllDrives"] = True

        try:
            self.metadata = (
                self.auth.service.files()
                .list(**dict(self))
                .execute(http=self.http)
            )
        except errors.HttpError as error:
            raise ApiRequestError(error)

        result = []
        for file_metadata in self.metadata["items"]:
            tmp_file = GoogleDriveFile(
                auth=self.auth, metadata=file_metadata, uploaded=True
            )
            result.append(tmp_file)
        return result


class IoBuffer:
    """Lightweight retention of one chunk."""

    def __init__(self, encoding):
        self.encoding = encoding
        self.chunk = None

    def write(self, chunk):
        self.chunk = chunk

    def read(self):
        return (
            self.chunk.decode(self.encoding)
            if self.chunk and self.encoding
            else self.chunk
        )


class MediaIoReadable:
    def __init__(
        self,
        request,
        encoding=None,
        pre_buffer=True,
        remove_prefix=b"",
        chunksize=DEFAULT_CHUNK_SIZE,
    ):
        """File-like wrapper around MediaIoBaseDownload.

        :param pre_buffer: Whether to read one chunk into an internal buffer
        immediately in order to raise any potential errors.
        :param remove_prefix: Bytes prefix to remove from internal pre_buffer.
        :raises: ApiRequestError
        """
        self.done = False
        self._fd = IoBuffer(encoding)
        self.downloader = MediaIoBaseDownload(
            self._fd, request, chunksize=chunksize
        )
        self.size = None
        self._pre_buffer = False
        if pre_buffer:
            self.read()
            if remove_prefix:
                chunk = io.BytesIO(self._fd.chunk)
                GoogleDriveFile._RemovePrefix(chunk, remove_prefix)
                self._fd.chunk = chunk.getvalue()
            self._pre_buffer = True

    def read(self):
        """
        :returns: bytes or str -- chunk (or None if done)
        :raises: ApiRequestError
        """
        if self._pre_buffer:
            self._pre_buffer = False
            return self._fd.read()
        if self.done:
            return None
        try:
            status, self.done = self.downloader.next_chunk()
            self.size = status.total_size
        except errors.HttpError as error:
            raise ApiRequestError(error)
        return self._fd.read()

    def __iter__(self):
        """
        :raises: ApiRequestError
        """
        while True:
            chunk = self.read()
            if chunk is None:
                break
            yield chunk

    def __len__(self):
        return self.size


class GoogleDriveFile(ApiAttributeMixin, ApiResource):
    """Google Drive File instance.

    Inherits ApiResource which inherits dict.
    Can access and modify metadata like dictionary.
    """

    content = ApiAttribute("content")
    uploaded = ApiAttribute("uploaded")
    metadata = ApiAttribute("metadata")

    def __init__(self, auth=None, metadata=None, uploaded=False):
        """Create an instance of GoogleDriveFile.

        :param auth: authorized GoogleAuth instance.
        :type auth: pydrive2.auth.GoogleAuth
        :param metadata: file resource to initialize GoogleDriveFile with.
        :type metadata: dict.
        :param uploaded: True if this file is confirmed to be uploaded.
        :type uploaded: bool.
        """
        ApiAttributeMixin.__init__(self)
        ApiResource.__init__(self)
        self.metadata = {}
        self.dirty = {"content": False}
        self.auth = auth
        self.uploaded = uploaded
        if uploaded:
            self.UpdateMetadata(metadata)
        elif metadata:
            self.update(metadata)
        self.has_bom = True

    def __getitem__(self, key):
        """Overwrites manner of accessing Files resource.

        If this file instance is not uploaded and id is specified,
        it will try to look for metadata with Files.get().

        :param key: key of dictionary query.
        :type key: str.
        :returns: value of Files resource
        :raises: KeyError, FileNotUploadedError
        """
        try:
            return dict.__getitem__(self, key)
        except KeyError as e:
            if self.uploaded:
                raise KeyError(e)
            if self.get("id"):
                self.FetchMetadata()
                return dict.__getitem__(self, key)
            else:
                raise FileNotUploadedError()

    def SetContentString(self, content, encoding="utf-8"):
        """Set content of this file to be a string.

        Creates io.BytesIO instance of utf-8 encoded string.
        Sets mimeType to be 'text/plain' if not specified and file id is not
        set (means that we are uploading this file for the first time).

        :param encoding: The encoding to use when setting the content of this file.
        :type encoding: str
        :param content: content of the file in string.
        :type content: str
        """
        self.content = io.BytesIO(content.encode(encoding))
        if self.get("mimeType") is None and self.get("id") is None:
            self["mimeType"] = "text/plain"

    def SetContentFile(self, filename):
        """Set content of this file from a file.

        Opens the file specified by this method.
        Will be read, uploaded, and closed by Upload() method.
        Sets metadata 'title' and 'mimeType' automatically if not specified and
        the file is uploaded for the first time (id is not set).

        :param filename: name of the file to be uploaded.
        :type filename: str.
        """
        self.content = open(filename, "rb")
        if self.get("id") is None:
            if self.get("title") is None:
                self["title"] = os.path.basename(filename)
            if self.get("mimeType") is None:
                self["mimeType"] = mimetypes.guess_type(filename)[0]

    def GetContentString(
        self, mimetype=None, encoding="utf-8", remove_bom=False
    ):
        """Get content of this file as a string.

        :param mimetype: The mimetype of the content string.
        :type mimetype: str

        :param encoding: The encoding to use when decoding the byte string.
        :type encoding: str

        :param remove_bom: Whether to strip a known BOM.
        :type remove_bom: bool

        :returns: str -- utf-8 decoded content of the file
        :raises: ApiRequestError, FileNotUploadedError, FileNotDownloadableError
        """
        if (
            self.content is None
            or not isinstance(self.content, io.BytesIO)
            or self.has_bom == remove_bom
        ):
            self.FetchContent(mimetype, remove_bom)
        return self.content.getvalue().decode(encoding)

    @LoadAuth
    def GetContentFile(
        self,
        filename,
        mimetype=None,
        remove_bom=False,
        callback=None,
        chunksize=DEFAULT_CHUNK_SIZE,
        acknowledge_abuse=False,
    ):
        """Save content of this file as a local file.

        :param filename: name of the file to write to.
        :type filename: str
        :param mimetype: mimeType of the file.
        :type mimetype: str
        :param remove_bom: Whether to remove the byte order marking.
        :type remove_bom: bool
        :param callback: passed two arguments: (total transferred, file size).
        :type param: callable
        :param chunksize: chunksize in bytes (standard 100 MB(1024*1024*100))
        :type chunksize: int
        :param acknowledge_abuse: Acknowledging the risk and download file
            identified as abusive.
        :type acknowledge_abuse: bool
        :raises: ApiRequestError, FileNotUploadedError
        """
        files = self.auth.service.files()
        file_id = self.metadata.get("id") or self.get("id")
        if not file_id:
            raise FileNotUploadedError()

        def download(fd, request):
            downloader = MediaIoBaseDownload(
                fd, self._WrapRequest(request), chunksize=chunksize
            )
            done = False
            while done is False:
                status, done = downloader.next_chunk()
                if callback:
                    callback(status.resumable_progress, status.total_size)

        with open(filename, mode="w+b") as fd:
            # Should use files.export_media instead of files.get_media if
            # metadata["mimeType"].startswith("application/vnd.google-apps.").
            # But that would first require a slow call to FetchMetadata().
            # We prefer to try-except for speed.
            try:
                download(
                    fd,
                    files.get_media(
                        fileId=file_id, acknowledgeAbuse=acknowledge_abuse
                    ),
                )
            except errors.HttpError as error:
                exc = ApiRequestError(error)
                if (
                    exc.error["code"] != 403
                    or exc.GetField("reason") != "fileNotDownloadable"
                ):
                    raise exc
                mimetype = mimetype or "text/plain"
                fd.seek(0)  # just in case `download()` modified `fd`
                try:
                    download(
                        fd,
                        files.export_media(fileId=file_id, mimeType=mimetype),
                    )
                except errors.HttpError as error:
                    raise ApiRequestError(error)

            if mimetype == "text/plain" and remove_bom:
                fd.seek(0)
                bom = self._GetBOM(mimetype)
                if bom:
                    self._RemovePrefix(fd, bom)

    @LoadAuth
    def GetContentIOBuffer(
        self,
        mimetype=None,
        encoding=None,
        remove_bom=False,
        chunksize=DEFAULT_CHUNK_SIZE,
        acknowledge_abuse=False,
    ):
        """Get a file-like object which has a buffered read() method.

        :param mimetype: mimeType of the file.
        :type mimetype: str
        :param encoding: The encoding to use when decoding the byte string.
        :type encoding: str
        :param remove_bom: Whether to remove the byte order marking.
        :type remove_bom: bool
        :param chunksize: default read()/iter() chunksize.
        :type chunksize: int
        :param acknowledge_abuse: Acknowledging the risk and download file
            identified as abusive.
        :type acknowledge_abuse: bool
        :returns: MediaIoReadable -- file-like object.
        :raises: ApiRequestError, FileNotUploadedError
        """
        files = self.auth.service.files()
        file_id = self.metadata.get("id") or self.get("id")
        if not file_id:
            raise FileNotUploadedError()

        # Should use files.export_media instead of files.get_media if
        # metadata["mimeType"].startswith("application/vnd.google-apps.").
        # But that would first require a slow call to FetchMetadata().
        # We prefer to try-except for speed.
        try:
            request = self._WrapRequest(
                files.get_media(
                    fileId=file_id, acknowledgeAbuse=acknowledge_abuse
                )
            )
            return MediaIoReadable(
                request, encoding=encoding, chunksize=chunksize
            )
        except ApiRequestError as exc:
            if (
                exc.error["code"] != 403
                or exc.GetField("reason") != "fileNotDownloadable"
            ):
                raise exc
            mimetype = mimetype or "text/plain"
            request = self._WrapRequest(
                files.export_media(fileId=file_id, mimeType=mimetype)
            )
            remove_prefix = (
                self._GetBOM(mimetype)
                if mimetype == "text/plain" and remove_bom
                else b""
            )
            return MediaIoReadable(
                request,
                encoding=encoding,
                remove_prefix=remove_prefix,
                chunksize=chunksize,
            )

    @LoadAuth
    def FetchMetadata(self, fields=None, fetch_all=False):
        """Download file's metadata from id using Files.get().

        :param fields: The fields to include, as one string, each entry separated
                       by commas, e.g. 'fields,labels'.
        :type fields: str

        :param fetch_all: Whether to fetch all fields.
        :type fetch_all: bool

        :raises: ApiRequestError, FileNotUploadedError
        """
        file_id = self.metadata.get("id") or self.get("id")

        if fetch_all:
            fields = "*"

        if file_id:
            try:
                request = self.auth.service.files().get(
                    fileId=file_id,
                    fields=fields,
                    # Teamdrive support
                    supportsAllDrives=True,
                )
                request = self._AddResourceKeyHeaders(request)
                metadata = request.execute(http=self.http)
            except errors.HttpError as error:
                raise ApiRequestError(error)
            else:
                self.uploaded = True
                self.UpdateMetadata(metadata)
        else:
            raise FileNotUploadedError()

    @LoadMetadata
    def FetchContent(self, mimetype=None, remove_bom=False):
        """Download file's content from download_url.

        :raises: ApiRequestError, FileNotUploadedError, FileNotDownloadableError
        """
        download_url = self.metadata.get("downloadUrl")
        export_links = self.metadata.get("exportLinks")
        if download_url:
            self.content = io.BytesIO(self._DownloadFromUrl(download_url))
            self.dirty["content"] = False

        elif export_links and export_links.get(mimetype):
            self.content = io.BytesIO(
                self._DownloadFromUrl(export_links.get(mimetype))
            )
            self.dirty["content"] = False

        else:
            raise FileNotDownloadableError(
                "No downloadLink/exportLinks for mimetype found in metadata"
            )

        if mimetype == "text/plain" and remove_bom:
            self._RemovePrefix(
                self.content, MIME_TYPE_TO_BOM[self["mimeType"]][mimetype]
            )
            self.has_bom = not remove_bom

    def Upload(self, param=None):
        """Upload/update file by choosing the most efficient method.

        :param param: additional parameter to upload file.
        :type param: dict.
        :raises: ApiRequestError
        """
        if self.uploaded or self.get("id") is not None:
            if self.dirty["content"]:
                self._FilesUpdate(param=param)
            else:
                self._FilesPatch(param=param)
        else:
            self._FilesInsert(param=param)

    def Trash(self, param=None):
        """Move a file to the trash.

        :raises: ApiRequestError
        """
        self._FilesTrash(param=param)

    def UnTrash(self, param=None):
        """Move a file out of the trash.
        :param param: Additional parameter to file.
        :type param: dict.
        :raises: ApiRequestError
        """
        self._FilesUnTrash(param=param)

    def Delete(self, param=None):
        """Hard-delete a file.

        :param param: additional parameter to file.
        :type param: dict.
        :raises: ApiRequestError
        """
        self._FilesDelete(param=param)

    @LoadAuth
    def Copy(self, target_folder=None, new_title=None, param=None):
        """Creates a copy of this file. Folders cannot be copied.

        :param target_folder: Folder where the file will be copied.
        :type target_folder: GoogleDriveFile, optional
        :param new_title: Name of the new file.
        :type new_title: str, optional
        :param param: addition parameters to pass.
        :type param: dict, optional
        :raises: ApiRequestError
        :return: the copied file
        :rtype: GoogleDriveFile
        """

        if param is None:
            param = {}

        param["fileId"] = self["id"]
        param["supportsAllDrives"] = True
        param["body"] = {}

        if target_folder:
            param["body"]["parents"] = [{"id": target_folder["id"]}]
        param["body"]["title"] = new_title

        new_file = None
        try:
            new_file = (
                self.auth.service.files().copy(**param).execute(http=self.http)
            )
        except errors.HttpError as error:
            raise ApiRequestError(error)

        return GoogleDriveFile(self.auth, new_file)

    def InsertPermission(self, new_permission, param=None):
        """Insert a new permission. Re-fetches all permissions after call.

        :param new_permission: The new permission to insert, please see the
                               official Google Drive API guide on permissions.insert
                               for details.
        :type new_permission: object

        :param param: addition parameters to pass
        :type param: dict

        :return: The permission object.
        :rtype: object
        """
        if param is None:
            param = {}
        param["fileId"] = self.metadata.get("id") or self["id"]
        param["body"] = new_permission
        # Teamdrive support
        param["supportsAllDrives"] = True

        try:
            permission = (
                self.auth.service.permissions()
                .insert(**param)
                .execute(http=self.http)
            )
        except errors.HttpError as error:
            raise ApiRequestError(error)
        else:
            self.GetPermissions()  # Update permissions field.

        return permission

    @LoadAuth
    def GetPermissions(self):
        """Get file's or shared drive's permissions.

        For files in a shared drive, at most 100 results will be returned.
        It doesn't paginate and collect all results.

        :return: A list of the permission objects.
        :rtype: object[]
        """
        file_id = self.metadata.get("id") or self.get("id")

        # We can't do FetchMetada call (which would nicely update
        # local metada cache, etc) here since it  doesn't return
        # permissions for the team drive use case.
        permissions = (
            self.auth.service.permissions()
            .list(
                fileId=file_id,
                # Teamdrive support
                supportsAllDrives=True,
            )
            .execute(http=self.http)
        ).get("items")

        if permissions:
            self["permissions"] = permissions
            self.metadata["permissions"] = permissions

        return permissions

    def DeletePermission(self, permission_id):
        """Deletes the permission specified by the permission_id.

        :param permission_id: The permission id.
        :type permission_id: str
        :return: True if it succeeds.
        :rtype: bool
        """
        return self._DeletePermission(permission_id)

    @LoadAuth
    def GetRevisions(self):
        """Get file's or shared drive's revisions.

        For files in a shared drive, at most 100 results will be returned.
        It doesn't paginate and collect all results.

        :return: A list of the revision objects.
        :rtype: object[]
        """
        file_id = self.metadata.get("id") or self.get("id")

        # We can't do FetchMetada call (which would nicely update
        # local metada cache, etc) here since it  doesn't return
        # revisions for the team drive use case.
        revisions = (
            self.auth.service.revisions()
            .list(
                fileId=file_id,
            )
            .execute(http=self.http)
        ).get("items")

        if revisions:
            self["revisions"] = revisions
            self.metadata["revisions"] = revisions

        return revisions

    def _WrapRequest(self, request):
        """Replaces request.http with self.http.

        Ensures thread safety. Similar to other places where we call
        `.execute(http=self.http)` to pass a client from the thread local storage.
        """
        if self.http:
            request.http = self.http
        request = self._AddResourceKeyHeaders(request)
        return request

    def _AddResourceKeyHeaders(self, request):
        """Add resourceKey headers to request if file is secured with resourceKey and
        its available (from a list for example).

        :param request: request to add headers to.
        :type request: googleapiclient.http.HttpRequest
        """
        file_id = self.metadata.get("id") or self.get("id")
        if file_id:
            resourceKey = self.get("resourceKey")
            if resourceKey:
                request.headers[
                    "X-Goog-Drive-Resource-Keys"
                ] = f"{file_id}/{resourceKey}"
        return request

    @LoadAuth
    def _FilesInsert(self, param=None):
        """Upload a new file using Files.insert().

        :param param: additional parameter to upload file.
        :type param: dict.
        :raises: ApiRequestError
        """
        if param is None:
            param = {}
        param["body"] = self.GetChanges()

        # teamdrive support
        param["supportsAllDrives"] = True

        try:
            if self.dirty["content"]:
                param["media_body"] = self._BuildMediaBody()
            metadata = (
                self.auth.service.files()
                .insert(**param)
                .execute(http=self.http)
            )
        except errors.HttpError as error:
            raise ApiRequestError(error)
        else:
            self.uploaded = True
            self.dirty["content"] = False
            self.UpdateMetadata(metadata)

    @LoadAuth
    def _FilesUnTrash(self, param=None):
        """Un-delete (Trash) a file using Files.UnTrash().
        :param param: additional parameter to file.
        :type param: dict.
        :raises: ApiRequestError
        """
        if param is None:
            param = {}
        param["fileId"] = self.metadata.get("id") or self["id"]

        # Teamdrive support
        param["supportsAllDrives"] = True

        try:
            self.auth.service.files().untrash(**param).execute(http=self.http)
        except errors.HttpError as error:
            raise ApiRequestError(error)
        else:
            if self.metadata:
                self.metadata["labels"]["trashed"] = False
            return True

    @LoadAuth
    def _FilesTrash(self, param=None):
        """Soft-delete (Trash) a file using Files.Trash().

        :param param: additional parameter to file.
        :type param: dict.
        :raises: ApiRequestError
        """
        if param is None:
            param = {}
        param["fileId"] = self.metadata.get("id") or self["id"]

        # Teamdrive support
        param["supportsAllDrives"] = True

        try:
            self.auth.service.files().trash(**param).execute(http=self.http)
        except errors.HttpError as error:
            raise ApiRequestError(error)
        else:
            if self.metadata:
                self.metadata["labels"]["trashed"] = True
            return True

    @LoadAuth
    def _FilesDelete(self, param=None):
        """Delete a file using Files.Delete()
        (WARNING: deleting permanently deletes the file!)

        :param param: additional parameter to file.
        :type param: dict.
        :raises: ApiRequestError
        """
        if param is None:
            param = {}
        param["fileId"] = self.metadata.get("id") or self["id"]

        # Teamdrive support
        param["supportsAllDrives"] = True

        try:
            self.auth.service.files().delete(**param).execute(http=self.http)
        except errors.HttpError as error:
            raise ApiRequestError(error)
        else:
            return True

    @LoadAuth
    def _FilesUpdate(self, param=None):
        """Update metadata and/or content using Files.Update().

        :param param: additional parameter to upload file.
        :type param: dict.
        :raises: ApiRequestError, FileNotUploadedError
        """
        if param is None:
            param = {}
        param["body"] = self.GetChanges()
        param["fileId"] = self.metadata.get("id") or self.get("id")

        # Teamdrive support
        param["supportsAllDrives"] = True

        try:
            if self.dirty["content"]:
                param["media_body"] = self._BuildMediaBody()
            metadata = (
                self.auth.service.files()
                .update(**param)
                .execute(http=self.http)
            )
        except errors.HttpError as error:
            raise ApiRequestError(error)
        else:
            self.uploaded = True
            self.dirty["content"] = False
            self.UpdateMetadata(metadata)

    @LoadAuth
    def _FilesPatch(self, param=None):
        """Update metadata using Files.Patch().

        :param param: additional parameter to upload file.
        :type param: dict.
        :raises: ApiRequestError, FileNotUploadedError
        """
        if param is None:
            param = {}
        param["body"] = self.GetChanges()
        param["fileId"] = self.metadata.get("id") or self.get("id")

        # Teamdrive support
        param["supportsAllDrives"] = True

        try:
            metadata = (
                self.auth.service.files()
                .patch(**param)
                .execute(http=self.http)
            )
        except errors.HttpError as error:
            raise ApiRequestError(error)
        else:
            self.UpdateMetadata(metadata)

    def _BuildMediaBody(self):
        """Build MediaIoBaseUpload to get prepared to upload content of the file.

        Sets mimeType as 'application/octet-stream' if not specified.

        :returns: MediaIoBaseUpload -- instance that will be used to upload content.
        """
        if self.get("mimeType") is None:
            self["mimeType"] = "application/octet-stream"
        return MediaIoBaseUpload(
            self.content, self["mimeType"], resumable=True
        )

    @LoadAuth
    def _DownloadFromUrl(self, url):
        """Download file from url using provided credential.

        :param url: link of the file to download.
        :type url: str.
        :returns: str -- content of downloaded file in string.
        :raises: ApiRequestError
        """
        resp, content = self.http.request(url)
        if resp.status != 200:
            raise ApiRequestError(errors.HttpError(resp, content, uri=url))
        return content

    @LoadAuth
    def _DeletePermission(self, permission_id):
        """Deletes the permission remotely, and from the file object itself.

        :param permission_id: The ID of the permission.
        :type permission_id: str

        :return: The permission
        :rtype: object
        """
        file_id = self.metadata.get("id") or self["id"]
        try:
            self.auth.service.permissions().delete(
                fileId=file_id, permissionId=permission_id
            ).execute()
        except errors.HttpError as error:
            raise ApiRequestError(error)
        else:
            if "permissions" in self and "permissions" in self.metadata:
                permissions = self["permissions"]
                is_not_current_permission = (
                    lambda per: per["id"] == permission_id
                )
                permissions = list(
                    filter(is_not_current_permission, permissions)
                )
                self["permissions"] = permissions
                self.metadata["permissions"] = permissions
            return True

    @staticmethod
    def _GetBOM(mimetype):
        """Based on download mime type (ignores Google Drive mime type)"""
        for bom in MIME_TYPE_TO_BOM.values():
            if mimetype in bom:
                return bom[mimetype]

    @staticmethod
    def _RemovePrefix(file_object, prefix, block_size=BLOCK_SIZE):
        """Deletes passed prefix by shifting content of passed file object by to
        the left. Operation is in-place.

        Args:
          file_object (obj): The file object to manipulate.
          prefix (str): The prefix to insert.
          block_size (int): The size of the blocks which are moved one at a time.
        """
        prefix_length = len(prefix)
        # Detect if prefix exists in file.
        content_start = file_object.read(prefix_length)

        if content_start == prefix:
            # Shift content left by prefix length, by copying 1KiB at a time.
            block_to_write = file_object.read(block_size)
            current_block_length = len(block_to_write)

            # Read and write location in separate variables for simplicity.
            read_location = prefix_length + current_block_length
            write_location = 0

            while current_block_length > 0:
                # Write next block.
                file_object.seek(write_location)
                file_object.write(block_to_write)
                # Set write location to the next block.
                write_location += len(block_to_write)

                # Read next block of input.
                file_object.seek(read_location)
                block_to_write = file_object.read(block_size)
                # Update the current block length and read_location.
                current_block_length = len(block_to_write)
                read_location += current_block_length

            # Truncate the file to its, now shorter, length.
            file_object.truncate(read_location - prefix_length)

    @staticmethod
    def _InsertPrefix(file_object, prefix, block_size=BLOCK_SIZE):
        """Inserts the passed prefix in the beginning of the file, operation is
        in-place.

        Args:
          file_object (obj): The file object to manipulate.
          prefix (str): The prefix to insert.
        """
        # Read the first two blocks.
        first_block = file_object.read(block_size)
        second_block = file_object.read(block_size)
        # Pointer to the first byte of the next block to be read.
        read_location = block_size * 2

        # Write BOM.
        file_object.seek(0)
        file_object.write(prefix)
        # {read|write}_location separated for readability.
        write_location = len(prefix)

        # Write and read block alternatingly.
        while len(first_block):
            # Write first block.
            file_object.seek(write_location)
            file_object.write(first_block)
            # Increment write_location.
            write_location += block_size

            # Move second block into first variable.
            first_block = second_block

            # Read in the next block.
            file_object.seek(read_location)
            second_block = file_object.read(block_size)
            # Increment read_location.
            read_location += block_size



from pathlib import Path
import random
import re
import os
import posixpath

from funcy import retry
from funcy.py3 import cat
from pydrive2.files import ApiRequestError
from shutil import copyfile, rmtree

newline_pattern = re.compile(r"[\r\n]")

GDRIVE_USER_CREDENTIALS_DATA = "GDRIVE_USER_CREDENTIALS_DATA"
DEFAULT_USER_CREDENTIALS_FILE = "/tmp/pydrive2/credentials.json"

TESTS_ROOTDIR = os.path.dirname(__file__)
SETTINGS_PATH = posixpath.join(TESTS_ROOTDIR, "settings/")
LOCAL_PATH = posixpath.join(TESTS_ROOTDIR, "settings/local/")


def setup_credentials(credentials_path=DEFAULT_USER_CREDENTIALS_FILE):
    os.chdir(TESTS_ROOTDIR)
    if os.getenv(GDRIVE_USER_CREDENTIALS_DATA):
        if not os.path.exists(os.path.dirname(credentials_path)):
            os.makedirs(os.path.dirname(credentials_path), exist_ok=True)
        with open(credentials_path, "w") as credentials_file:
            credentials_file.write(os.getenv(GDRIVE_USER_CREDENTIALS_DATA))


def settings_file_path(settings_file, wkdir=LOCAL_PATH):
    template_path = SETTINGS_PATH + settings_file
    wkdir = Path(wkdir)
    local_path = wkdir / settings_file
    assert os.path.exists(template_path)
    if not os.path.exists(wkdir):
        os.makedirs(wkdir, exist_ok=True)
    if not os.path.exists(local_path):
        copyfile(template_path, local_path)
    return local_path


class PyDriveRetriableError(Exception):
    pass


# 15 tries, start at 0.5s, multiply by golden ratio, cap at 20s
@retry(15, PyDriveRetriableError, timeout=lambda a: min(0.5 * 1.618**a, 20))
def pydrive_retry(call, *args, **kwargs):
    try:
        result = call(*args, **kwargs)
    except ApiRequestError as exception:
        if exception.error["code"] in [403, 500, 502, 503, 504]:
            raise PyDriveRetriableError("Google API request failed")
        raise
    return result


def pydrive_list_item(drive, query, max_results=1000):
    param = {"q": query, "maxResults": max_results}

    file_list = drive.ListFile(param)

    # Isolate and decorate fetching of remote drive items in pages
    get_list = lambda: pydrive_retry(next, file_list, None)  # noqa: E731

    # Fetch pages until None is received, lazily flatten the thing
    return cat(iter(get_list, None))


def CreateRandomFileName():
    hash = random.getrandbits(128)
    return "%032x" % hash


def StripNewlines(string):
    return newline_pattern.sub("", string)


def create_file(path, content):
    with open(path, "w") as f:
        f.write(content)


def delete_file(path):
    if os.path.exists(path):
        os.remove(path)
        return True
    return False


def delete_dir(path):
    rmtree(path, ignore_errors=True)



import unittest

from pydrive2.auth import GoogleAuth
from pydrive2.drive import GoogleDrive
from pydrive2.test.test_util import (
    pydrive_retry,
    setup_credentials,
    settings_file_path,
)


class GoogleDriveTest(unittest.TestCase):
    """Tests basic operations on meta-data information of the linked Google
    Drive account.
    """

    @classmethod
    def setup_class(cls):
        setup_credentials()

        cls.ga = GoogleAuth(settings_file_path("default.yaml"))
        cls.ga.ServiceAuth()

    def test_01_About_Request(self):
        drive = GoogleDrive(self.ga)

        about_object = pydrive_retry(drive.GetAbout)
        self.assertTrue(about_object is not None, "About object not loading.")


if __name__ == "__main__":
    unittest.main()



import filecmp
import json
import os
import unittest
from unittest.mock import MagicMock
import httplib2
import pytest
import sys
from io import BytesIO
from tempfile import mkdtemp
from time import time

import timeout_decorator
from concurrent.futures import ThreadPoolExecutor, as_completed
from googleapiclient import errors

from pydrive2.auth import GoogleAuth
from pydrive2.drive import GoogleDrive
from pydrive2.files import ApiRequestError, GoogleDriveFile
from pydrive2.test import test_util
from pydrive2.test.test_util import (
    pydrive_retry,
    setup_credentials,
    create_file,
    delete_dir,
    delete_file,
    settings_file_path,
)


def auth_with_resource_key_mock() -> GoogleAuth:
    """
    Create GoogleAuth with mocked inner httplib2.Http simulating need
    for resourceKey header.
    """
    http_mock = MagicMock()

    def resource_key_request(
        uri,
        method="GET",
        body=None,
        headers=None,
        redirections=1,
        connection_type=None,
    ):
        """httplib2.Http.request mock."""
        # If "media" is not requested, it means we expect metadata
        fetch_meta_data = "alt=media" not in uri
        if (
            headers
            and "X-Goog-Drive-Resource-Keys" in headers
            and headers["X-Goog-Drive-Resource-Keys"]
            == "0BxphPoRgwhnodHNjS3JESnFNS1E/0-vjzOveuin3fnf4LUlfsD3A"
        ):
            if fetch_meta_data:
                # Fake meta data query response
                content = json.dumps({"title": "N48E012.zip"}).encode()
            else:
                # Fake file content query response
                content = b"some content"
            return (
                httplib2.Response(
                    {"status": "200", "content-length": str(len(content))}
                ),
                content,
            )
        # Simulate 404 response for file not found; body must be valid error JSON
        return (
            httplib2.Response({"status": "404"}),
            json.dumps({"error": {"code": 404}}).encode(),
        )

    http_mock.request.side_effect = resource_key_request
    ga = GoogleAuth(
        settings_file_path(
            "default.yaml", os.path.join(os.path.dirname(__file__), "")
        )
    )
    ga.thread_local.http = http_mock
    ga.ServiceAuth()
    return ga


class GoogleDriveFileTest(unittest.TestCase):
    """Tests basic file operations of files.GoogleDriveFile.
    Upload and download of contents and metadata, and thread-safety checks.
    Equivalent to Files.insert, Files.update, Files.patch in Google Drive API.
    """

    @classmethod
    def setup_class(cls):
        setup_credentials()

        cls.tmpdir = mkdtemp()

        cls.ga = GoogleAuth(
            settings_file_path("default.yaml", os.path.join(cls.tmpdir, ""))
        )
        cls.ga.ServiceAuth()

    @classmethod
    def tearDownClass(cls):
        delete_dir(cls.tmpdir)

    @classmethod
    def getTempFile(cls, prefix="", content=""):
        filename = os.path.join(cls.tmpdir, prefix + str(time()))
        if content:
            create_file(filename, content)
        return filename

    def test_Files_Insert(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("firsttestfile")
        file1["title"] = filename
        pydrive_retry(file1.Upload)  # Files.insert

        self.assertEqual(file1.metadata["title"], filename)
        file2 = drive.CreateFile({"id": file1["id"]})  # Download file from id.
        self.assertEqual(file2["title"], filename)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Insert_Unicode(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("첫번째 파일")
        file1["title"] = filename
        pydrive_retry(file1.Upload)  # Files.insert

        self.assertEqual(file1.metadata["title"], filename)
        file2 = drive.CreateFile({"id": file1["id"]})  # Download file from id.
        self.assertEqual(file2["title"], filename)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Insert_Content_String(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("secondtestfile")
        content = "hello world!"
        file1["title"] = filename
        file1.SetContentString(content)
        pydrive_retry(file1.Upload)  # Files.insert

        self.assertEqual(file1.GetContentString(), content)

        pydrive_retry(
            file1.FetchContent
        )  # Force download and double check content
        self.assertEqual(file1.metadata["title"], filename)
        self.assertEqual(file1.GetContentString(), content)

        file2 = drive.CreateFile({"id": file1["id"]})  # Download file from id.
        pydrive_retry(file2.FetchContent)
        self.assertEqual(file2.GetContentString(), content)
        self.assertEqual(file2.metadata["title"], filename)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Insert_Content_Unicode_String(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("두번째 파일")
        content = "안녕 세상아!"
        file1["title"] = filename
        file1.SetContentString(content)
        pydrive_retry(file1.Upload)  # Files.insert

        self.assertEqual(file1.GetContentString(), content)
        self.assertEqual(file1.metadata["title"], filename)
        pydrive_retry(
            file1.FetchContent
        )  # Force download and double check content.
        self.assertEqual(file1.GetContentString(), content)

        file2 = drive.CreateFile({"id": file1["id"]})  # Download file from id.
        pydrive_retry(file2.FetchContent)
        self.assertEqual(file2.GetContentString(), content)
        self.assertEqual(file2.metadata["title"], filename)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Insert_Content_File(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        contentFile = self.getTempFile("actual_content", "some string")
        file1.SetContentFile(contentFile)
        pydrive_retry(file1.Upload)  # Files.insert

        self.assertEqual(
            file1.metadata["title"], os.path.basename(contentFile)
        )
        pydrive_retry(
            file1.FetchContent
        )  # Force download and double check content.
        fileOut = self.getTempFile()
        pydrive_retry(file1.GetContentFile, fileOut)
        self.assertEqual(filecmp.cmp(contentFile, fileOut), True)

        file2 = drive.CreateFile({"id": file1["id"]})  # Download file from id.
        fileOut = self.getTempFile()
        pydrive_retry(file2.GetContentFile, fileOut)
        self.assertEqual(filecmp.cmp(contentFile, fileOut), True)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Patch(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("prepatchtestfile")
        newfilename = self.getTempFile("patchtestfile")
        file1["title"] = filename
        pydrive_retry(file1.Upload)  # Files.insert

        self.assertEqual(file1.metadata["title"], filename)
        file1["title"] = newfilename
        pydrive_retry(file1.Upload)  # Files.patch

        self.assertEqual(file1.metadata["title"], newfilename)
        file2 = drive.CreateFile({"id": file1["id"]})  # Download file from id.
        pydrive_retry(file2.FetchMetadata)
        self.assertEqual(file2.metadata["title"], newfilename)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Patch_By_Id(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("prepatchtestfile")
        newfilename = self.getTempFile("patchtestfile")
        file1["title"] = filename
        pydrive_retry(file1.Upload)  # Files.insert

        file2 = drive.CreateFile(
            {"id": file1["id"]}
        )  # Patch file no download.
        file2["title"] = newfilename
        pydrive_retry(file2.Upload)  # Files.patch

        file3 = drive.CreateFile({"id": file1["id"]})  # Download file from id.
        pydrive_retry(file3.FetchMetadata)
        self.assertEqual(file3.metadata["title"], newfilename)

    def test_Files_Patch_Skipping_Content(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("prepatchtestfile")
        newfilename = self.getTempFile("patchtestfile")
        content = "hello world!"

        file1["title"] = filename
        file1.SetContentString(content)
        pydrive_retry(file1.Upload)  # Files.insert
        self.assertEqual(file1.metadata["title"], filename)

        file1["title"] = newfilename
        pydrive_retry(file1.Upload)  # Files.patch
        self.assertEqual(file1.metadata["title"], newfilename)
        self.assertEqual(file1.GetContentString(), content)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Update_String(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("preupdatetestfile")
        newfilename = self.getTempFile("updatetestfile")
        content = "hello world!"
        newcontent = "hello new world!"

        file1["title"] = filename
        file1.SetContentString(content)
        pydrive_retry(file1.Upload)  # Files.insert
        self.assertEqual(file1.metadata["title"], filename)
        self.assertEqual(file1.GetContentString(), content)

        pydrive_retry(
            file1.FetchContent
        )  # Force download and double check content.
        self.assertEqual(file1.GetContentString(), content)

        file1["title"] = newfilename
        file1.SetContentString(newcontent)
        pydrive_retry(file1.Upload)  # Files.update
        self.assertEqual(file1.metadata["title"], newfilename)
        self.assertEqual(file1.GetContentString(), newcontent)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Update_File(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("preupdatetestfile")
        newfilename = self.getTempFile("updatetestfile")
        contentFile = self.getTempFile("actual_content", "some string")
        contentFile2 = self.getTempFile("actual_content_2", "some string")

        file1["title"] = filename
        file1.SetContentFile(contentFile)
        pydrive_retry(file1.Upload)  # Files.insert
        self.assertEqual(file1.metadata["title"], filename)

        pydrive_retry(
            file1.FetchContent
        )  # Force download and double check content.
        fileOut = self.getTempFile()
        pydrive_retry(file1.GetContentFile, fileOut)
        self.assertEqual(filecmp.cmp(contentFile, fileOut), True)

        file1["title"] = newfilename
        file1.SetContentFile(contentFile2)
        pydrive_retry(file1.Upload)  # Files.update
        self.assertEqual(file1.metadata["title"], newfilename)

        fileOut = self.getTempFile()
        pydrive_retry(file1.GetContentFile, fileOut)
        self.assertEqual(filecmp.cmp(contentFile2, fileOut), True)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Update_File_Keeps_Filename(self):
        # Tests https://github.com/iterative/PyDrive2/issues/272
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("preupdatetestfile")
        contentFile = self.getTempFile("actual_content", "some string")
        contentFile2 = self.getTempFile("actual_content_2", "some string")

        file1["title"] = filename
        file1.SetContentFile(contentFile)
        pydrive_retry(file1.Upload)  # Files.insert
        self.assertEqual(file1.metadata["title"], filename)

        same_file = drive.CreateFile({"id": file1["id"]})
        same_file.SetContentFile(contentFile2)
        pydrive_retry(same_file.Upload)  # Files.update
        self.assertEqual(same_file.metadata["title"], filename)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Update_By_Id(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("preupdatetestfile")
        newfilename = self.getTempFile("updatetestfile")
        content = "hello world!"
        newcontent = "hello new world!"

        file1["title"] = filename
        file1.SetContentString(content)
        pydrive_retry(file1.Upload)  # Files.insert

        file2 = drive.CreateFile({"id": file1["id"]})
        file2["title"] = newfilename
        file2.SetContentString(newcontent)
        pydrive_retry(file2.Upload)  # Files.update

        file3 = drive.CreateFile({"id": file1["id"]})
        pydrive_retry(file3.FetchContent)
        self.assertEqual(file3.GetContentString(), newcontent)
        self.assertEqual(file3["title"], newfilename)

    def test_Files_Download_Service(self):
        """
        Tests that a fresh GoogleDrive object can correctly authenticate
        and download from a file ID.
        """
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile("prepatchtestfile")
        content = "hello world!"

        file1["title"] = filename
        file1.SetContentString(content)
        pydrive_retry(file1.Upload)  # Files.insert
        self.assertEqual(file1.metadata["title"], filename)
        fileOut1 = self.getTempFile()
        pydrive_retry(file1.GetContentFile, fileOut1)

        # fresh download-only instance
        auth = GoogleAuth(
            settings_file_path("default.yaml", os.path.join(self.tmpdir, ""))
        )
        auth.ServiceAuth()
        drive2 = GoogleDrive(auth)
        file2 = drive2.CreateFile({"id": file1["id"]})
        fileOut2 = self.getTempFile()
        pydrive_retry(file2.GetContentFile, fileOut2)
        self.assertEqual(filecmp.cmp(fileOut1, fileOut2), True)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Get_Content_Buffer(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        filename = self.getTempFile()
        content = "hello world!\ngoodbye, cruel world!"
        file1["title"] = filename
        file1.SetContentString(content)
        pydrive_retry(file1.Upload)  # Files.insert

        buffer1 = pydrive_retry(file1.GetContentIOBuffer)
        self.assertEqual(file1.metadata["title"], filename)
        self.assertEqual(len(buffer1), len(content))
        self.assertEqual(b"".join(iter(buffer1)).decode("ascii"), content)

        buffer2 = pydrive_retry(file1.GetContentIOBuffer, encoding="ascii")
        self.assertEqual(len(buffer2), len(content))
        self.assertEqual("".join(iter(buffer2)), content)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Get_Content_Buffer_resourceKey_missing(self):
        """404 expected for file secured with resourceKey when not provided."""

        ga = auth_with_resource_key_mock()

        drive = GoogleDrive(ga)
        file1 = drive.CreateFile(
            {
                "id": "0BxphPoRgwhnodHNjS3JESnFNS1E",
            }
        )
        with self.assertRaisesRegex(
            ApiRequestError, "HttpError 404 when requesting"
        ):
            pydrive_retry(file1.GetContentIOBuffer)

    def test_Files_Get_Content_Buffer_resourceKey(self):
        """End to end scenario with real file."""
        ga = auth_with_resource_key_mock()
        drive = GoogleDrive(ga)
        file1 = drive.CreateFile(
            {
                "id": "0BxphPoRgwhnodHNjS3JESnFNS1E",
                "resourceKey": "0-vjzOveuin3fnf4LUlfsD3A",
            }
        )

        buffer1 = pydrive_retry(file1.GetContentIOBuffer)

        self.assertEqual(len(buffer1), 12)

    @pytest.mark.manual
    def test_Files_Get_Content_Buffer_resourceKey_missing_real(self):
        """
        404 expected for file secured with resourceKey when not provided.
        End to end scenario with real public file.
        """
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile(
            {
                "id": "0BxphPoRgwhnodHNjS3JESnFNS1E",
            }
        )
        with self.assertRaisesRegex(
            ApiRequestError, "HttpError 404 when requesting"
        ):
            pydrive_retry(file1.GetContentIOBuffer)

    @pytest.mark.manual
    def test_Files_Get_Content_Buffer_resourceKey_real(self):
        """End to end scenario with real public file."""
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile(
            {
                "id": "0BxphPoRgwhnodHNjS3JESnFNS1E",
                "resourceKey": "0-vjzOveuin3fnf4LUlfsD3A",
            }
        )

        buffer1 = pydrive_retry(file1.GetContentIOBuffer)

        self.assertEqual(len(buffer1), 6128902)

    def test_Files_Get_Content_File_resourceKey_missing(self):
        """404 expected for file secured with resourceKey when not provided."""
        ga = auth_with_resource_key_mock()
        drive = GoogleDrive(ga)
        file1 = drive.CreateFile(
            {
                "id": "0BxphPoRgwhnodHNjS3JESnFNS1E",
            }
        )
        fileOut = self.getTempFile()
        with self.assertRaisesRegex(
            ApiRequestError, "HttpError 404 when requesting"
        ):
            pydrive_retry(file1.GetContentFile, fileOut)

    def test_Files_Get_Content_File_resourceKey(self):
        ga = auth_with_resource_key_mock()
        drive = GoogleDrive(ga)
        file1 = drive.CreateFile(
            {
                "id": "0BxphPoRgwhnodHNjS3JESnFNS1E",
                "resourceKey": "0-vjzOveuin3fnf4LUlfsD3A",
            }
        )

        fileOut = self.getTempFile()
        pydrive_retry(file1.GetContentFile, fileOut)

        with open(fileOut, "rb") as f:
            self.assertEqual(len(f.read()), 12)

    def test_Files_Fetch_Metadata_resourceKey_missing(self):
        """404 expected for file secured with resourceKey when not provided."""
        ga = auth_with_resource_key_mock()
        drive = GoogleDrive(ga)
        file1 = drive.CreateFile(
            {
                "id": "0BxphPoRgwhnodHNjS3JESnFNS1E",
            }
        )
        with self.assertRaisesRegex(
            ApiRequestError, "HttpError 404 when requesting"
        ):
            pydrive_retry(file1.FetchMetadata)

    def test_Files_Fetch_Metadata_resourceKey(self):
        ga = auth_with_resource_key_mock()
        drive = GoogleDrive(ga)
        file1 = drive.CreateFile(
            {
                "id": "0BxphPoRgwhnodHNjS3JESnFNS1E",
                "resourceKey": "0-vjzOveuin3fnf4LUlfsD3A",
            }
        )

        pydrive_retry(file1.FetchMetadata)

        self.assertEqual(file1.metadata["title"], "N48E012.zip")

    def test_Upload_Download_Empty_File(self):
        filename = os.path.join(self.tmpdir, str(time()))
        create_file(filename, "")

        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        file1.SetContentFile(filename)
        pydrive_retry(file1.Upload)

        fileOut1 = self.getTempFile()
        pydrive_retry(file1.GetContentFile, fileOut1)
        self.assertEqual(os.path.getsize(fileOut1), 0)

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Upload_Download_Empty_String(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        file1.SetContentString("")
        pydrive_retry(file1.Upload)

        self.assertEqual(pydrive_retry(file1.GetContentString), "")

        # Force download and double check content
        pydrive_retry(file1.FetchContent)
        self.assertEqual(file1.GetContentString(), "")

        # Download file from id
        file2 = drive.CreateFile({"id": file1["id"]})
        pydrive_retry(file2.FetchContent)
        self.assertEqual(file2.GetContentString(), "")

        self.DeleteUploadedFiles(drive, [file1["id"]])

    # Tests for Trash/UnTrash/Delete.
    # ===============================

    def test_Files_Trash_File(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)
        self.assertFalse(file1.metadata["labels"]["trashed"])

        # Download to verify non-trashed state on GDrive.
        file2 = drive.CreateFile({"id": file1["id"]})
        pydrive_retry(file2.FetchMetadata)
        self.assertFalse(file2.metadata["labels"]["trashed"])

        pydrive_retry(file1.Trash)
        self.assertTrue(file1.metadata["labels"]["trashed"])

        pydrive_retry(file2.FetchMetadata)
        self.assertTrue(file2.metadata["labels"]["trashed"])

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Trash_File_Just_ID(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)
        self.assertFalse(file1.metadata["labels"]["trashed"])

        # Trash file by ID.
        file2 = drive.CreateFile({"id": file1["id"]})
        pydrive_retry(file2.Trash)

        # Verify trashed by downloading metadata.
        pydrive_retry(file1.FetchMetadata)
        self.assertTrue(file1.metadata["labels"]["trashed"])

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_UnTrash_File(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)
        pydrive_retry(file1.Trash)
        self.assertTrue(file1.metadata["labels"]["trashed"])

        # Verify that file is trashed by downloading metadata.
        file2 = drive.CreateFile({"id": file1["id"]})
        pydrive_retry(file2.FetchMetadata)
        self.assertTrue(file2.metadata["labels"]["trashed"])

        # Un-trash the file, and assert local metadata is updated correctly.
        pydrive_retry(file1.UnTrash)
        self.assertFalse(file1.metadata["labels"]["trashed"])

        # Re-fetch the metadata, and assert file un-trashed on GDrive.
        pydrive_retry(file2.FetchMetadata)
        self.assertFalse(file2.metadata["labels"]["trashed"])

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_UnTrash_File_Just_ID(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)
        pydrive_retry(file1.Trash)
        self.assertTrue(file1.metadata["labels"]["trashed"])

        file2 = drive.CreateFile({"id": file1["id"]})
        pydrive_retry(file2.UnTrash)  # UnTrash without fetching metadata.

        pydrive_retry(file1.FetchMetadata)
        self.assertFalse(file1.metadata["labels"]["trashed"])

        self.DeleteUploadedFiles(drive, [file1["id"]])

    def test_Files_Delete_File(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)
        file2 = drive.CreateFile({"id": file1["id"]})

        pydrive_retry(file1.Delete)

        try:
            pydrive_retry(file2.FetchMetadata)
            self.fail("File not deleted correctly.")
        except ApiRequestError:
            pass

    def test_Files_Delete_File_Just_ID(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)
        file2 = drive.CreateFile({"id": file1["id"]})

        pydrive_retry(file2.Delete)

        try:
            pydrive_retry(file1.FetchMetadata)
            self.fail("File not deleted correctly.")
        except ApiRequestError:
            pass

    # Tests for Permissions.
    # ======================

    def test_Files_FetchMetadata_Fields(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)

        self.assertFalse("permissions" in file1)

        pydrive_retry(file1.FetchMetadata, "permissions")
        self.assertTrue("permissions" in file1)
        pydrive_retry(file1.Delete)

    def test_Files_FetchAllMetadata_Fields(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)

        pydrive_retry(file1.FetchMetadata, fetch_all=True)
        self.assertTrue("hasThumbnail" in file1)
        self.assertTrue("thumbnailVersion" in file1)
        self.assertTrue("permissions" in file1)
        pydrive_retry(file1.Delete)

    def test_Files_Insert_Permission(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)

        # Verify only one permission before inserting permission.
        permissions = pydrive_retry(file1.GetPermissions)
        self.assertEqual(len(permissions), 1)
        self.assertEqual(len(file1["permissions"]), 1)

        # Insert the permission.
        permission = pydrive_retry(
            file1.InsertPermission,
            {"type": "anyone", "value": "anyone", "role": "reader"},
        )
        self.assertTrue(permission)
        self.assertEqual(len(file1["permissions"]), 2)
        self.assertEqual(file1["permissions"][0]["type"], "anyone")

        permissions = pydrive_retry(file1.GetPermissions)
        self.assertEqual(len(file1["permissions"]), 2)
        self.assertEqual(file1["permissions"][0]["type"], "anyone")
        self.assertEqual(permissions[0]["type"], "anyone")

        # Verify remote changes made.
        file2 = drive.CreateFile({"id": file1["id"]})
        permissions = pydrive_retry(file2.GetPermissions)
        self.assertEqual(len(permissions), 2)
        self.assertEqual(permissions[0]["type"], "anyone")

        pydrive_retry(file1.Delete)

    def test_Files_Get_Permissions(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)

        self.assertFalse("permissions" in file1)

        permissions = pydrive_retry(file1.GetPermissions)
        self.assertTrue(permissions is not None)
        self.assertTrue("permissions" in file1)

        pydrive_retry(file1.Delete)

    def test_Files_Delete_Permission(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)
        pydrive_retry(
            file1.InsertPermission,
            {"type": "anyone", "value": "anyone", "role": "reader"},
        )
        permissions = pydrive_retry(file1.GetPermissions)
        self.assertEqual(len(permissions), 2)
        self.assertEqual(len(file1["permissions"]), 2)

        pydrive_retry(file1.DeletePermission, permissions[0]["id"])
        self.assertEqual(len(file1["permissions"]), 1)

        # Verify remote changes made.
        file2 = drive.CreateFile({"id": file1["id"]})
        permissions = pydrive_retry(file2.GetPermissions)
        self.assertEqual(len(permissions), 1)

        pydrive_retry(file1.Delete)

    def test_Files_Delete_Permission_Invalid(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)

        try:
            pydrive_retry(file1.DeletePermission, "invalid id")
            self.fail("Deleting invalid permission not raising exception.")
        except ApiRequestError:
            pass

        pydrive_retry(file1.Delete)

    def test_Files_Get_Revisions(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()
        pydrive_retry(file1.Upload)

        self.assertFalse("revisions" in file1)

        revisions = pydrive_retry(file1.GetRevisions)
        self.assertTrue(revisions is not None)
        self.assertTrue("revisions" in file1)

        pydrive_retry(file1.Delete)

    def test_ApiRequestError_HttpError_Propagation(self):
        file = GoogleDrive(self.ga).CreateFile()
        pydrive_retry(file.Upload)
        try:
            pydrive_retry(file.DeletePermission, "invalid id")
            self.fail("Deleting invalid permission not raising exception.")
        except ApiRequestError as exc:
            self.assertTrue(
                exc.args and isinstance(exc.args[0], errors.HttpError)
            )
            self.assertTrue(exc.error is not None)
            # Validating for HttpError 404 "Permission not found: invalid id"
            self.assertTrue(exc.error["code"] == 404)
        finally:
            pydrive_retry(file.Delete)

    def test_GFile_Conversion_Lossless_String(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()

        # Upload a string, and convert into Google Doc format.
        test_string = "Generic, non-exhaustive ASCII test string."
        file1.SetContentString(test_string)
        pydrive_retry(file1.Upload, {"convert": True})

        # Download string as plain text.
        downloaded_string = file1.GetContentString(mimetype="text/plain")
        self.assertEqual(
            test_string, downloaded_string, "Strings do not match"
        )

        # Download content into file and ensure that file content matches original
        # content string.
        downloaded_file_name = "_tmp_downloaded_file_name.txt"
        pydrive_retry(
            file1.GetContentFile,
            downloaded_file_name,
            mimetype="text/plain",
            remove_bom=True,
        )
        downloaded_string = open(downloaded_file_name).read()
        self.assertEqual(
            test_string, downloaded_string, "Strings do not match"
        )

        # Delete temp file.
        delete_file(downloaded_file_name)

    # Tests for GDrive conversion.
    # ============================

    def setup_gfile_conversion_test(self):
        drive = GoogleDrive(self.ga)
        file1 = drive.CreateFile()

        # Create a file to upload.
        file_name = "_tmp_source_file.txt"
        downloaded_file_name = "_tmp_downloaded_file_name.txt"
        original_file_content = "Generic, non-exhaustive\n ASCII test string."
        source_file = open(file_name, mode="w+")
        source_file.write(original_file_content)
        source_file.close()
        original_file_content = test_util.StripNewlines(original_file_content)

        return file1, file_name, original_file_content, downloaded_file_name

    def cleanup_gfile_conversion_test(
        self, file1, file_name, downloaded_file_name
    ):
        # Delete temporary files.
        os.path.exists(file_name) and os.remove(file_name)
        os.path.exists(downloaded_file_name) and os.remove(
            downloaded_file_name
        )
        pydrive_retry(file1.Delete)  # Delete uploaded file.

    def test_GFile_Conversion_Remove_BOM(self):
        (
            file1,
            file_name,
            original_file_content,
            downloaded_file_name,
        ) = self.setup_gfile_conversion_test()
        try:
            # Upload source_file and convert into Google Doc format.
            file1.SetContentFile(file_name)
            pydrive_retry(file1.Upload, {"convert": True})

            # Download as string.
            downloaded_content_no_bom = file1.GetContentString(
                mimetype="text/plain", remove_bom=True
            )
            downloaded_content_no_bom = test_util.StripNewlines(
                downloaded_content_no_bom
            )
            self.assertEqual(original_file_content, downloaded_content_no_bom)

            # Download as file.
            pydrive_retry(
                file1.GetContentFile, downloaded_file_name, remove_bom=True
            )
            downloaded_content = open(downloaded_file_name).read()
            downloaded_content = test_util.StripNewlines(downloaded_content)
            self.assertEqual(original_file_content, downloaded_content)

        finally:
            self.cleanup_gfile_conversion_test(
                file1, file_name, downloaded_file_name
            )

    def test_Gfile_Conversion_Add_Remove_BOM(self):
        """Tests whether you can switch between the BOM appended and removed
        version on the fly."""
        (
            file1,
            file_name,
            original_file_content,
            downloaded_file_name,
        ) = self.setup_gfile_conversion_test()
        try:
            file1.SetContentFile(file_name)
            pydrive_retry(file1.Upload, {"convert": True})

            content_bom = file1.GetContentString(mimetype="text/plain")
            content_no_bom = file1.GetContentString(
                mimetype="text/plain", remove_bom=True
            )
            content_bom_2 = file1.GetContentString(mimetype="text/plain")

            self.assertEqual(content_bom, content_bom_2)
            self.assertNotEqual(content_bom, content_no_bom)
            self.assertTrue(len(content_bom) > len(content_no_bom))

            buffer_bom = pydrive_retry(
                file1.GetContentIOBuffer,
                mimetype="text/plain",
                encoding="utf-8",
            )
            buffer_bom = "".join(iter(buffer_bom))
            self.assertEqual(content_bom, buffer_bom)

            buffer_no_bom = pydrive_retry(
                file1.GetContentIOBuffer,
                mimetype="text/plain",
                remove_bom=True,
                encoding="utf-8",
            )
            buffer_no_bom = "".join(iter(buffer_no_bom))
            self.assertEqual(content_no_bom, buffer_no_bom)

        finally:
            self.cleanup_gfile_conversion_test(
                file1, file_name, downloaded_file_name
            )

    def test_InsertPrefix(self):
        # Create BytesIO.
        file_obj = BytesIO(b"abc")
        original_length = len(file_obj.getvalue())
        char_to_insert = "\ufeff".encode()

        # Insert the prefix.
        GoogleDriveFile._InsertPrefix(file_obj, char_to_insert)
        modified_length = len(file_obj.getvalue())
        self.assertGreater(modified_length, original_length)
        self.assertEqual(file_obj.getvalue(), "\ufeffabc".encode())

    def test_InsertPrefixLarge(self):
        # Create BytesIO.
        test_content = "abc" * 800
        file_obj = BytesIO(test_content.encode("utf-8"))
        original_length = len(file_obj.getvalue())
        char_to_insert = "\ufeff".encode()

        # Insert the prefix.
        GoogleDriveFile._InsertPrefix(file_obj, char_to_insert)
        modified_length = len(file_obj.getvalue())
        self.assertGreater(modified_length, original_length)
        expected_content = "\ufeff" + test_content
        self.assertEqual(file_obj.getvalue(), expected_content.encode("utf8"))

    def test_RemovePrefix(self):
        # Create BytesIO.
        file_obj = BytesIO("\ufeffabc".encode())
        original_length = len(file_obj.getvalue())
        char_to_remove = "\ufeff".encode()

        # Insert the prefix.
        GoogleDriveFile._RemovePrefix(file_obj, char_to_remove)
        modified_length = len(file_obj.getvalue())
        self.assertLess(modified_length, original_length)
        self.assertEqual(file_obj.getvalue(), b"abc")

    def test_RemovePrefixLarge(self):
        # Create BytesIO.
        test_content = "\ufeff" + "abc" * 800
        file_obj = BytesIO(test_content.encode("utf8"))
        original_length = len(file_obj.getvalue())
        char_to_remove = "\ufeff".encode()

        # Insert the prefix.
        GoogleDriveFile._RemovePrefix(file_obj, char_to_remove)
        modified_length = len(file_obj.getvalue())
        self.assertLess(modified_length, original_length)
        self.assertEqual(file_obj.getvalue(), test_content[1:].encode("utf8"))

    # Setup for concurrent upload testing.
    # =====================================
    FILE_UPLOAD_COUNT = 10

    def _parallel_uploader(self, num_of_uploads, num_of_workers):
        """
        :returns: list[str] of file IDs
        """
        drive = GoogleDrive(self.ga)
        thread_pool = ThreadPoolExecutor(max_workers=num_of_workers)
        first_file = self.getTempFile("first_file", "some string")
        second_file = self.getTempFile("second_file", "another string")

        # Create list of gdrive_files.
        upload_files = []
        remote_name = test_util.CreateRandomFileName()
        for i in range(num_of_uploads):
            file_name = first_file if i % 2 == 0 else second_file
            up_file = drive.CreateFile()
            up_file["title"] = remote_name
            up_file.SetContentFile(file_name)
            upload_files.append(up_file)

        # Ensure there are no files with the random file name.
        files = pydrive_retry(
            lambda: drive.ListFile(
                param={"q": "title = '%s' and trashed = false" % remote_name}
            ).GetList()
        )
        self.assertTrue(len(files) == 0)

        # Submit upload jobs to ThreadPoolExecutor.
        futures = []
        for up_file in upload_files:
            futures.append(thread_pool.submit(pydrive_retry, up_file.Upload))

        # Ensure that all threads a) return, and b) encountered no exceptions.
        for future in as_completed(futures):
            self.assertIsNone(future.exception())
        thread_pool.shutdown()

        # Ensure all files were uploaded.
        files = pydrive_retry(
            lambda: drive.ListFile(
                param={"q": "title = '%s' and trashed = false" % remote_name}
            ).GetList()
        )
        self.assertTrue(len(files) == self.FILE_UPLOAD_COUNT)

        return [fi["id"] for fi in upload_files]

    def _parallel_downloader(self, file_ids, num_of_workers):
        drive = GoogleDrive(self.ga)
        thread_pool = ThreadPoolExecutor(max_workers=num_of_workers)

        # Create list of gdrive_files.
        download_files = []
        for file_id in file_ids:
            file1 = drive.CreateFile({"id": file_id})
            file1["title"] = self.getTempFile()
            download_files.append(file1)

        # Ensure files don't exist yet.
        for file_obj in download_files:
            self.assertTrue(not delete_file(file_obj["title"]))

        # Submit upload jobs to ThreadPoolExecutor.
        futures = []
        for file_obj in download_files:
            futures.append(
                thread_pool.submit(
                    pydrive_retry, file_obj.GetContentFile, file_obj["title"]
                )
            )

        # Ensure that all threads a) return, and b) encountered no exceptions.
        for future in as_completed(futures):
            self.assertIsNone(future.exception())
        thread_pool.shutdown()

        # Ensure all files were downloaded.
        for file_obj in download_files:
            self.assertTrue(delete_file(file_obj["title"]))

        # Remove uploaded files.
        self.DeleteUploadedFiles(drive, file_ids)

    @pytest.mark.skipif(
        sys.platform == "win32",
        reason="timeout_decorator doesn't support Windows",
    )
    @timeout_decorator.timeout(320)
    def test_Parallel_Insert_File_Passed_HTTP(self):
        files = self._parallel_uploader(self.FILE_UPLOAD_COUNT, 10)
        self._parallel_downloader(files, 10)

    # Tests for Copy file.
    # ====================

    def test_CopyFileSameFolder(self):
        drive = GoogleDrive(self.ga)

        content = "hello world!"

        # create a temp file and set it's content to a known string
        file1 = drive.CreateFile()
        filename = self.getTempFile("copytestfile", content=content)
        file1["title"] = filename
        file1.SetContentFile(filename)
        pydrive_retry(file1.Upload)

        # copy the file
        file2 = pydrive_retry(file1.Copy, new_title="copytestfile_copy")

        self.assertIsNotNone(file2)

        pydrive_retry(file2.FetchContent)
        # assert that the content of the copied file is the same as the original (file1)
        self.assertEqual(file2.GetContentString(), content)

        self.DeleteUploadedFiles(drive, [file1["id"], file2["id"]])

    def test_CopyFileDifferentFolder(self):
        drive = GoogleDrive(self.ga)

        content = "hello world!"

        # create a temp file and set it's content to a known string
        file1 = drive.CreateFile()
        filename = self.getTempFile("copytestfile", content=content)
        file1["title"] = filename
        file1.SetContentFile(filename)
        pydrive_retry(file1.Upload)

        # create a temp directory
        temp_dir = pydrive_retry(
            drive.CreateFile,
            {
                "title": "temp_dir",
                "mimeType": "application/vnd.google-apps.folder",
                "parents": [{"id": file1["parents"][0]["id"]}],
            },
        )
        pydrive_retry(temp_dir.Upload)

        # copy the file into the new folder
        file2 = pydrive_retry(
            file1.Copy, target_folder=temp_dir, new_title="copytestfile_copy"
        )

        self.assertIsNotNone(file2)

        pydrive_retry(file2.FetchContent)
        # assert that the content of the copied file is the same as the original (file1)
        self.assertEqual(file2.GetContentString(), content)

        files = pydrive_retry(
            drive.ListFile, {"q": f"'{temp_dir['id']}' in parents"}
        ).GetList()
        self.assertIn("copytestfile_copy", [f["title"] for f in files])

        self.DeleteUploadedFiles(
            drive, [file1["id"], file2["id"], temp_dir["id"]]
        )

    # Helper functions.
    # =================

    def DeleteUploadedFiles(self, drive, ids):
        for element in ids:
            tmp_file = drive.CreateFile({"id": element})
            pydrive_retry(tmp_file.Delete)


if __name__ == "__main__":
    unittest.main()



import json
import os
import re
import time
import pytest

from pydrive2.auth import AuthenticationError, GoogleAuth
from pydrive2.test.test_util import (
    setup_credentials,
    delete_file,
    settings_file_path,
    GDRIVE_USER_CREDENTIALS_DATA,
)
from oauth2client.file import Storage


def setup_module(module):
    setup_credentials()


@pytest.mark.manual
def test_01_LocalWebserverAuthWithClientConfigFromFile():
    # Delete old credentials file
    delete_file("credentials/1.dat")
    # Test if authentication works with config read from file
    ga = GoogleAuth(settings_file_path("test_oauth_test_01.yaml"))
    ga.LocalWebserverAuth()
    assert not ga.access_token_expired
    # Test if correct credentials file is created
    CheckCredentialsFile("credentials/1.dat")
    time.sleep(1)


@pytest.mark.manual
def test_02_LocalWebserverAuthWithClientConfigFromSettings():
    # Delete old credentials file
    delete_file("credentials/2.dat")
    # Test if authentication works with config read from settings
    ga = GoogleAuth(settings_file_path("test_oauth_test_02.yaml"))
    ga.LocalWebserverAuth()
    assert not ga.access_token_expired
    # Test if correct credentials file is created
    CheckCredentialsFile("credentials/2.dat")
    time.sleep(1)


@pytest.mark.manual
def test_03_LocalWebServerAuthWithNoCredentialsSaving():
    # Delete old credentials file
    delete_file("credentials/3.dat")
    ga = GoogleAuth(settings_file_path("test_oauth_test_03.yaml"))
    assert not ga.settings["save_credentials"]
    ga.LocalWebserverAuth()
    assert not ga.access_token_expired
    time.sleep(1)


@pytest.mark.manual
def test_04_CommandLineAuthWithClientConfigFromFile():
    # Delete old credentials file
    delete_file("credentials/4.dat")
    # Test if authentication works with config read from file
    ga = GoogleAuth(settings_file_path("test_oauth_test_04.yaml"))
    ga.CommandLineAuth()
    assert not ga.access_token_expired
    # Test if correct credentials file is created
    CheckCredentialsFile("credentials/4.dat")
    time.sleep(1)


@pytest.mark.manual
def test_05_ConfigFromSettingsWithoutOauthScope():
    # Test if authentication works without oauth_scope
    ga = GoogleAuth(settings_file_path("test_oauth_test_05.yaml"))
    ga.LocalWebserverAuth()
    assert not ga.access_token_expired
    time.sleep(1)


@pytest.mark.skip(reason="P12 authentication is deprecated")
def test_06_ServiceAuthFromSavedCredentialsP12File():
    setup_credentials("credentials/6.dat")
    ga = GoogleAuth(settings_file_path("test_oauth_test_06.yaml"))
    ga.ServiceAuth()
    assert not ga.access_token_expired
    time.sleep(1)


def test_07_ServiceAuthFromSavedCredentialsJsonFile():
    # Have an initial auth so that credentials/7.dat gets saved
    ga = GoogleAuth(settings_file_path("test_oauth_test_07.yaml"))
    credentials_file = ga.settings["save_credentials_file"]
    # Delete old credentials file
    delete_file(credentials_file)
    assert not os.path.exists(credentials_file)
    ga.ServiceAuth()
    assert os.path.exists(credentials_file)
    # Secondary auth should be made only using the previously saved
    # login info
    ga = GoogleAuth(settings_file_path("test_oauth_test_07.yaml"))
    ga.ServiceAuth()
    assert not ga.access_token_expired
    time.sleep(1)


def test_08_ServiceAuthFromJsonFileNoCredentialsSaving():
    # Test that no credentials are saved and API is still functional
    # We are testing that there are no exceptions at least
    ga = GoogleAuth(settings_file_path("test_oauth_test_08.yaml"))
    assert not ga.settings["save_credentials"]
    ga.ServiceAuth()
    time.sleep(1)


def test_09_SaveLoadCredentialsUsesDefaultStorage(mocker):
    # Test fix for https://github.com/iterative/PyDrive2/issues/163
    # Make sure that Load and Save credentials by default reuse the
    # same Storage (since it defined lock which make it TS)
    ga = GoogleAuth(settings_file_path("test_oauth_test_09.yaml"))
    credentials_file = ga.settings["save_credentials_file"]
    # Delete old credentials file
    delete_file(credentials_file)
    assert not os.path.exists(credentials_file)
    spy = mocker.spy(Storage, "__init__")
    ga.ServiceAuth()
    ga.LoadCredentials()
    ga.SaveCredentials()
    assert spy.call_count == 0


def test_10_ServiceAuthFromSavedCredentialsDictionary():
    creds_dict = {}
    settings = {
        "client_config_backend": "service",
        "service_config": {
            "client_json_file_path": "/tmp/pydrive2/credentials.json",
        },
        "oauth_scope": ["https://www.googleapis.com/auth/drive"],
        "save_credentials": True,
        "save_credentials_backend": "dictionary",
        "save_credentials_dict": creds_dict,
        "save_credentials_key": "creds",
    }
    ga = GoogleAuth(settings=settings)
    ga.ServiceAuth()
    assert not ga.access_token_expired
    assert creds_dict
    first_creds_dict = creds_dict.copy()
    # Secondary auth should be made only using the previously saved
    # login info
    ga = GoogleAuth(settings=settings)
    ga.ServiceAuth()
    assert not ga.access_token_expired
    assert creds_dict == first_creds_dict
    time.sleep(1)


def test_11_ServiceAuthFromJsonNoCredentialsSaving():
    client_json = os.environ[GDRIVE_USER_CREDENTIALS_DATA]
    settings = {
        "client_config_backend": "service",
        "service_config": {
            "client_json": client_json,
        },
        "oauth_scope": ["https://www.googleapis.com/auth/drive"],
    }
    # Test that no credentials are saved and API is still functional
    # We are testing that there are no exceptions at least
    ga = GoogleAuth(settings=settings)
    assert not ga.settings["save_credentials"]
    ga.ServiceAuth()
    time.sleep(1)


def test_12_ServiceAuthFromJsonDictNoCredentialsSaving():
    client_json_dict = json.loads(os.environ[GDRIVE_USER_CREDENTIALS_DATA])
    settings = {
        "client_config_backend": "service",
        "service_config": {
            "client_json_dict": client_json_dict,
        },
        "oauth_scope": ["https://www.googleapis.com/auth/drive"],
    }
    # Test that no credentials are saved and API is still functional
    # We are testing that there are no exceptions at least
    ga = GoogleAuth(settings=settings)
    assert not ga.settings["save_credentials"]
    ga.ServiceAuth()
    time.sleep(1)


def test_13_LocalWebServerAuthNonInterativeRaises(monkeypatch):
    settings = {
        "client_config_backend": "file",
        "client_config_file": "client_secrets.json",
        "oauth_scope": ["https://www.googleapis.com/auth/drive"],
    }
    ga = GoogleAuth(settings=settings)

    monkeypatch.setenv("GDRIVE_NON_INTERACTIVE", "true")
    # Test that exception is raised on trying to do browser auth if
    # we are running in a non interactive environment.
    with pytest.raises(
        AuthenticationError,
        match=re.escape(
            "Non interactive mode (GDRIVE_NON_INTERACTIVE env) is enabled"
        ),
    ):
        ga.LocalWebserverAuth()


def CheckCredentialsFile(credentials, no_file=False):
    ga = GoogleAuth(settings_file_path("test_oauth_default.yaml"))
    ga.LoadCredentialsFile(credentials)
    assert ga.access_token_expired == no_file



import appdirs
import errno
import io
import logging
import os
import posixpath
import threading
from collections import defaultdict
from contextlib import contextmanager
from itertools import chain

from fsspec.spec import AbstractFileSystem
from funcy import cached_property, retry, wrap_prop, wrap_with
from tqdm.utils import CallbackIOWrapper

from pydrive2.drive import GoogleDrive
from pydrive2.fs.utils import IterStream
from pydrive2.auth import GoogleAuth

logger = logging.getLogger(__name__)

FOLDER_MIME_TYPE = "application/vnd.google-apps.folder"

COMMON_SETTINGS = {
    "get_refresh_token": True,
    "oauth_scope": [
        "https://www.googleapis.com/auth/drive",
        "https://www.googleapis.com/auth/drive.appdata",
    ],
}


class GDriveAuthError(Exception):
    pass


def _gdrive_retry(func):
    def should_retry(exc):
        from pydrive2.files import ApiRequestError

        if not isinstance(exc, ApiRequestError):
            return False

        error_code = exc.error.get("code", 0)
        result = False
        if 500 <= error_code < 600:
            result = True

        if error_code == 403:
            result = exc.GetField("reason") in [
                "userRateLimitExceeded",
                "rateLimitExceeded",
            ]
        if result:
            logger.debug(f"Retrying GDrive API call, error: {exc}.")

        return result

    # 16 tries, start at 0.5s, multiply by golden ratio, cap at 20s
    return retry(
        16,
        timeout=lambda a: min(0.5 * 1.618**a, 20),
        filter_errors=should_retry,
    )(func)


@contextmanager
def _wrap_errors():
    try:
        yield
    except Exception as exc:
        # Handle AuthenticationError, RefreshError and other auth failures
        # It's hard to come up with a narrow exception, since PyDrive throws
        # a lot of different errors - broken credentials file, refresh token
        # expired, flow failed, etc.
        raise GDriveAuthError("Failed to authenticate GDrive") from exc


def _client_auth(
    client_id=None,
    client_secret=None,
    client_json=None,
    client_json_file_path=None,
    profile=None,
):
    if client_json:
        save_settings = {
            "save_credentials_backend": "dictionary",
            "save_credentials_dict": {"creds": client_json},
            "save_credentials_key": "creds",
        }
    else:
        creds_file = client_json_file_path
        if not creds_file:
            cache_dir = os.path.join(
                appdirs.user_cache_dir("pydrive2fs", appauthor=False),
                client_id,
            )
            os.makedirs(cache_dir, exist_ok=True)

            profile = profile or "default"
            creds_file = os.path.join(cache_dir, f"{profile}.json")

        save_settings = {
            "save_credentials_backend": "file",
            "save_credentials_file": creds_file,
        }

    settings = {
        **COMMON_SETTINGS,
        "save_credentials": True,
        **save_settings,
        "client_config_backend": "settings",
        "client_config": {
            "client_id": client_id,
            "client_secret": client_secret,
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "revoke_uri": "https://oauth2.googleapis.com/revoke",
            "redirect_uri": "",
        },
    }

    auth = GoogleAuth(settings=settings)

    with _wrap_errors():
        auth.LocalWebserverAuth()

    return auth


def _service_auth(
    client_user_email=None,
    client_json=None,
    client_json_file_path=None,
):
    settings = {
        **COMMON_SETTINGS,
        "client_config_backend": "service",
        "service_config": {
            "client_user_email": client_user_email,
            "client_json": client_json,
            "client_json_file_path": client_json_file_path,
        },
    }

    auth = GoogleAuth(settings=settings)

    with _wrap_errors():
        auth.ServiceAuth()

    return auth


class GDriveFileSystem(AbstractFileSystem):
    """Access to gdrive as an fsspec filesystem"""

    def __init__(
        self,
        path,
        google_auth=None,
        client_id=None,
        client_secret=None,
        client_user_email=None,
        client_json=None,
        client_json_file_path=None,
        use_service_account=False,
        profile=None,
        trash_only=True,
        acknowledge_abuse=False,
        **kwargs,
    ):
        """Create an instance of GDriveFileSystem.

        :param path: gdrive path.
        :type path: str.
        :param google_auth: Authenticated GoogleAuth instance.
        :type google_auth: GoogleAuth.
        :param client_id: Client ID of the application.
        :type client_id: str
        :param client_secret: Client secret of the application.
        :type client_secret: str.
        :param client_user_email: User email that authority was delegated to
            (only for service account).
        :type client_user_email: str.
        :param client_json: JSON keyfile loaded into a string.
        :type client_json: str.
        :param client_json_file_path: Path to JSON keyfile.
        :type client_json_file_path: str.
        :param use_service_account: Use service account.
        :type use_service_account: bool.
        :param profile: Profile name for caching credentials
            (ignored for service account).
        :param trash_only: Move files to trash instead of deleting.
        :type trash_only: bool.
        :param acknowledge_abuse: Acknowledging the risk and download file
            identified as abusive.
        :type acknowledge_abuse: bool
        :type profile: str.
        :raises: GDriveAuthError
        """
        super().__init__(**kwargs)
        self.path = path
        self.root, self.base = self.split_path(self.path)

        if not google_auth:
            if (
                not client_json
                and not client_json_file_path
                and not (client_id and client_secret)
            ):
                raise ValueError(
                    "Specify credentials using one of these methods: "
                    "client_id/client_secret or "
                    "client_json or "
                    "client_json_file_path"
                )

            if use_service_account:
                google_auth = _service_auth(
                    client_json=client_json,
                    client_json_file_path=client_json_file_path,
                    client_user_email=client_user_email,
                )
            else:
                google_auth = _client_auth(
                    client_id=client_id,
                    client_secret=client_secret,
                    client_json=client_json,
                    client_json_file_path=client_json_file_path,
                    profile=profile,
                )

        self.client = GoogleDrive(google_auth)
        self._trash_only = trash_only
        self._acknowledge_abuse = acknowledge_abuse

    def split_path(self, path):
        parts = path.replace("//", "/").rstrip("/").split("/", 1)
        if len(parts) == 2:
            return parts
        else:
            return parts[0], ""

    @wrap_prop(threading.RLock())
    @cached_property
    def _ids_cache(self):
        cache = {"dirs": defaultdict(list), "ids": {}}

        base_item_ids = self._path_to_item_ids(self.base, use_cache=False)
        if not base_item_ids:
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                f"Confirm {self.path} exists and you can access it",
            )

        self._cache_path_id(self.base, *base_item_ids, cache=cache)

        return cache

    def _cache_path_id(self, path, *item_ids, cache=None):
        cache = cache or self._ids_cache
        for item_id in item_ids:
            cache["ids"][item_id] = path
            cache["dirs"][path].append(item_id)

    @cached_property
    def _list_params(self):
        params = {"corpora": "default"}
        if self.root != "root" and self.root != "appDataFolder":
            drive_id = self._gdrive_shared_drive_id(self.root)
            if drive_id:
                logger.debug(
                    "GDrive remote '{}' is using shared drive id '{}'.".format(
                        self.path, drive_id
                    )
                )
                params["driveId"] = drive_id
                params["corpora"] = "drive"
        return params

    @_gdrive_retry
    def _gdrive_shared_drive_id(self, item_id):
        from pydrive2.files import ApiRequestError

        param = {"id": item_id}
        # it does not create a file on the remote
        item = self.client.CreateFile(param)
        # ID of the shared drive the item resides in.
        # Only populated for items in shared drives.
        try:
            item.FetchMetadata("driveId")
        except ApiRequestError as exc:
            error_code = exc.error.get("code", 0)
            if error_code == 404:
                raise PermissionError from exc
            raise

        return item.get("driveId", None)

    def _gdrive_list(self, query):
        param = {"q": query, "maxResults": 1000}
        param.update(self._list_params)
        file_list = self.client.ListFile(param)

        # Isolate and decorate fetching of remote drive items in pages.
        get_list = _gdrive_retry(lambda: next(file_list, None))

        # Fetch pages until None is received, lazily flatten the thing.
        return chain.from_iterable(iter(get_list, None))

    def _gdrive_list_ids(self, query_ids):
        query = " or ".join(
            f"'{query_id}' in parents" for query_id in query_ids
        )
        query = f"({query}) and trashed=false"
        return self._gdrive_list(query)

    def _get_remote_item_ids(
        self, parent_ids, parent_path, title, use_cache=True
    ):
        if not parent_ids:
            return None
        query = "trashed=false and ({})".format(
            " or ".join(
                f"'{parent_id}' in parents" for parent_id in parent_ids
            )
        )
        query += " and title='{}'".format(title.replace("'", "\\'"))

        res = []
        for item in self._gdrive_list(query):
            # GDrive list API is case insensitive, we need to compare
            # all results and pick the ones with the right title
            if item["title"] == title:
                res.append(item["id"])

                if item["mimeType"] == FOLDER_MIME_TYPE and use_cache:
                    self._cache_path_id(
                        posixpath.join(parent_path, item["title"]), item["id"]
                    )

        return res

    def _get_cached_item_ids(self, path, use_cache):
        if not path:
            return [self.root]
        if use_cache:
            return self._ids_cache["dirs"].get(path, [])
        return []

    def _path_to_item_ids(self, path, create=False, use_cache=True):
        item_ids = self._get_cached_item_ids(path, use_cache)
        if item_ids:
            return item_ids

        parent_path, title = posixpath.split(path)
        parent_ids = self._path_to_item_ids(parent_path, create, use_cache)
        item_ids = self._get_remote_item_ids(
            parent_ids, parent_path, title, use_cache
        )
        if item_ids:
            return item_ids

        return (
            [self._create_dir(min(parent_ids), title, path)] if create else []
        )

    def _get_item_id(self, path, create=False, use_cache=True):
        bucket, base = self.split_path(path)
        assert bucket == self.root

        item_ids = self._path_to_item_ids(base, create, use_cache)
        if item_ids:
            return min(item_ids)

        assert not create
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

    @_gdrive_retry
    def _gdrive_create_dir(self, parent_id, title):
        parent = {"id": parent_id}
        item = self.client.CreateFile(
            {"title": title, "parents": [parent], "mimeType": FOLDER_MIME_TYPE}
        )
        item.Upload()
        return item

    @wrap_with(threading.RLock())
    def _create_dir(self, parent_id, title, remote_path):
        cached = self._ids_cache["dirs"].get(remote_path)
        if cached:
            return cached[0]

        item = self._gdrive_create_dir(parent_id, title)
        self._cache_path_id(remote_path, item["id"])

        return item["id"]

    def exists(self, path):
        try:
            self._get_item_id(path)
        except FileNotFoundError:
            return False
        else:
            return True

    @_gdrive_retry
    def info(self, path):
        bucket, base = self.split_path(path)
        item_id = self._get_item_id(path)
        gdrive_file = self.client.CreateFile({"id": item_id})
        gdrive_file.FetchMetadata()

        metadata = {"name": posixpath.join(bucket, base.rstrip("/"))}
        if gdrive_file["mimeType"] == FOLDER_MIME_TYPE:
            metadata["type"] = "directory"
            metadata["size"] = 0
            metadata["name"] += "/"
        else:
            metadata["type"] = "file"
            metadata["size"] = int(gdrive_file.get("fileSize"))
            metadata["checksum"] = gdrive_file.get("md5Checksum")
        return metadata

    def ls(self, path, detail=False):
        bucket, base = self.split_path(path)
        assert bucket == self.root

        dir_ids = self._path_to_item_ids(base)
        if not dir_ids:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), path
            )

        root_path = posixpath.join(bucket, base)
        contents = []
        for item in self._gdrive_list_ids(dir_ids):
            item_path = posixpath.join(root_path, item["title"])
            if item["mimeType"] == FOLDER_MIME_TYPE:
                contents.append(
                    {
                        "type": "directory",
                        "name": item_path.rstrip("/") + "/",
                        "size": 0,
                    }
                )
            else:
                size = item.get("fileSize")
                contents.append(
                    {
                        "type": "file",
                        "name": item_path,
                        "size": int(size) if size is not None else size,
                        "checksum": item.get("md5Checksum"),
                    }
                )

        if detail:
            return contents
        else:
            return [content["name"] for content in contents]

    def find(self, path, detail=False, **kwargs):
        bucket, base = self.split_path(path)
        assert bucket == self.root

        # Make sure the base path is cached and dir_ids below has some
        # dirs revelant to this call
        self._path_to_item_ids(base)

        dir_ids = [self._ids_cache["ids"].copy()]
        seen_paths = set()
        contents = []
        while dir_ids:
            query_ids = {
                dir_id: dir_name
                for dir_id, dir_name in dir_ids.pop().items()
                if posixpath.commonpath([base, dir_name]) == base
                if dir_id not in seen_paths
            }
            if not query_ids:
                continue

            seen_paths |= query_ids.keys()

            new_query_ids = {}
            dir_ids.append(new_query_ids)
            for item in self._gdrive_list_ids(query_ids):
                parent_id = item["parents"][0]["id"]
                item_path = posixpath.join(query_ids[parent_id], item["title"])
                if item["mimeType"] == FOLDER_MIME_TYPE:
                    new_query_ids[item["id"]] = item_path
                    self._cache_path_id(item_path, item["id"])
                    continue
                size = item.get("fileSize")
                contents.append(
                    {
                        "name": posixpath.join(bucket, item_path),
                        "type": "file",
                        "size": int(size) if size is not None else size,
                        "checksum": item.get("md5Checksum"),
                    }
                )

        if detail:
            return {content["name"]: content for content in contents}
        else:
            return [content["name"] for content in contents]

    def upload_fobj(self, stream, rpath, callback=None, **kwargs):
        parent_id = self._get_item_id(self._parent(rpath), create=True)
        if callback:
            stream = CallbackIOWrapper(
                callback.relative_update, stream, "read"
            )
        return self._gdrive_upload_fobj(
            posixpath.basename(rpath.rstrip("/")), parent_id, stream
        )

    def put_file(self, lpath, rpath, callback=None, **kwargs):
        if callback:
            callback.set_size(os.path.getsize(lpath))
        with open(lpath, "rb") as stream:
            self.upload_fobj(stream, rpath, callback=callback)

    @_gdrive_retry
    def _gdrive_upload_fobj(self, title, parent_id, stream, callback=None):
        item = self.client.CreateFile(
            {"title": title, "parents": [{"id": parent_id}]}
        )
        item.content = stream
        item.Upload()
        return item

    def cp_file(self, lpath, rpath, **kwargs):
        """In-memory streamed copy"""
        with self.open(lpath) as stream:
            # IterStream objects doesn't support full-length
            # seek() calls, so we have to wrap the data with
            # an external buffer.
            buffer = io.BytesIO(stream.read())
            self.upload_fobj(buffer, rpath)

    @_gdrive_retry
    def mv(self, path1, path2, maxdepth=None, **kwargs):
        if maxdepth is not None:
            raise NotImplementedError("Max depth move is not supported")

        src_name = posixpath.basename(path1)

        src_parent = self._parent(path1)

        if self.exists(path2):
            dst_name = src_name
            dst_parent = path2
        else:
            dst_name = posixpath.basename(path2)
            dst_parent = self._parent(path2)

        file1_id = self._get_item_id(path1)

        file1 = self.client.CreateFile({"id": file1_id})

        if src_name != dst_name:
            file1["title"] = dst_name

        if src_parent != dst_parent:
            file2_parent_id = self._get_item_id(dst_parent)
            file1["parents"] = [{"id": file2_parent_id}]

        # TODO need to invalidate the cache for the old path, see #232
        file1.Upload()

    def get_file(self, rpath, lpath, callback=None, block_size=None, **kwargs):
        item_id = self._get_item_id(rpath)
        return self._gdrive_get_file(
            item_id, lpath, callback=callback, block_size=block_size
        )

    @_gdrive_retry
    def _gdrive_get_file(self, item_id, rpath, callback=None, block_size=None):
        param = {"id": item_id}
        # it does not create a file on the remote
        gdrive_file = self.client.CreateFile(param)

        extra_args = {"acknowledge_abuse": self._acknowledge_abuse}
        if block_size:
            extra_args["chunksize"] = block_size

        if callback:

            def cb(value, _):
                callback.absolute_update(value)

            gdrive_file.FetchMetadata(fields="fileSize")
            callback.set_size(int(gdrive_file.get("fileSize")))
            extra_args["callback"] = cb

        gdrive_file.GetContentFile(rpath, **extra_args)

    def _open(self, path, mode, **kwargs):
        assert mode in {"rb", "wb"}
        if mode == "wb":
            return GDriveBufferedWriter(self, path)
        else:
            item_id = self._get_item_id(path)
            return self._gdrive_open_file(item_id)

    @_gdrive_retry
    def _gdrive_open_file(self, item_id):
        param = {"id": item_id}
        # it does not create a file on the remote
        gdrive_file = self.client.CreateFile(param)
        fd = gdrive_file.GetContentIOBuffer(
            acknowledge_abuse=self._acknowledge_abuse
        )
        return IterStream(iter(fd))

    def rm_file(self, path):
        item_id = self._get_item_id(path)
        self._gdrive_delete_file(item_id)

    @_gdrive_retry
    def _gdrive_delete_file(self, item_id):
        from pydrive2.files import ApiRequestError

        param = {"id": item_id}
        # it does not create a file on the remote
        item = self.client.CreateFile(param)

        try:
            item.Trash() if self._trash_only else item.Delete()
        except ApiRequestError as exc:
            http_error_code = exc.error.get("code", 0)
            if (
                http_error_code == 403
                and self._list_params["corpora"] == "drive"
                and exc.GetField("location") == "file.permissions"
            ):
                raise PermissionError(
                    "Insufficient permissions to {}. You should have {} "
                    "access level for the used shared drive. More details "
                    "at {}.".format(
                        "move the file into Trash"
                        if self._trash_only
                        else "permanently delete the file",
                        "Manager or Content Manager"
                        if self._trash_only
                        else "Manager",
                        "https://support.google.com/a/answer/7337554",
                    )
                ) from exc
            raise


class GDriveBufferedWriter(io.IOBase):
    def __init__(self, fs, path):
        self.fs = fs
        self.path = path
        self.buffer = io.BytesIO()
        self._closed = False

    def write(self, *args, **kwargs):
        self.buffer.write(*args, **kwargs)

    def readable(self):
        return False

    def writable(self):
        return not self.readable()

    def flush(self):
        self.buffer.flush()
        try:
            self.fs.upload_fobj(self.buffer, self.path)
        finally:
            self._closed = True

    def close(self):
        if self._closed:
            return None

        self.flush()
        self.buffer.close()
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    @property
    def closed(self):
        return self._closed



import os
import unittest

from pydrive2.auth import GoogleAuth
from pydrive2.drive import GoogleDrive
from pydrive2.test import test_util
from pydrive2.test.test_util import (
    pydrive_retry,
    pydrive_list_item,
    setup_credentials,
    settings_file_path,
)


class GoogleDriveFileListTest(unittest.TestCase):
    """Tests operations of files.GoogleDriveFileList class.
    Equivalent to Files.list in Google Drive API.
    """

    @classmethod
    def setup_class(cls):
        setup_credentials()

        cls.ga = GoogleAuth(settings_file_path("default.yaml"))
        cls.ga.ServiceAuth()
        cls.drive = GoogleDrive(cls.ga)

    def test_01_Files_List_GetList(self):
        drive = GoogleDrive(self.ga)
        query = f"title = '{self.title}' and trashed = false"
        for file1 in pydrive_list_item(drive, query):
            found = False
            for file2 in pydrive_list_item(drive, query):
                if file1["id"] == file2["id"]:
                    found = True
            self.assertEqual(found, True)

    def test_02_Files_List_ForLoop(self):
        drive = GoogleDrive(self.ga)
        query = f"title = '{self.title}' and trashed = false"
        files = []
        for x in pydrive_list_item(
            drive, query, 2
        ):  # Build iterator to access files simply with for loop
            files.append(x)
        for file1 in self.file_list:
            found = False
            for file2 in files:
                if file1["id"] == file2["id"]:
                    found = True
            self.assertEqual(found, True)

    def test_03_Files_List_GetList_Iterate(self):
        drive = GoogleDrive(self.ga)
        flist = drive.ListFile(
            {
                "q": "title = '%s' and trashed = false" % self.title,
                "maxResults": 2,
            }
        )
        files = []
        while True:
            try:
                x = pydrive_retry(flist.GetList)
                self.assertTrue(len(x) <= 2)
                files.extend(x)
            except StopIteration:
                break
        for file1 in self.file_list:
            found = False
            for file2 in files:
                if file1["id"] == file2["id"]:
                    found = True
            self.assertEqual(found, True)

    def test_File_List_Folders(self):
        drive = GoogleDrive(self.ga)
        folder1 = drive.CreateFile(
            {
                "mimeType": "application/vnd.google-apps.folder",
                "title": self.title,
            }
        )
        pydrive_retry(folder1.Upload)
        self.file_list.append(folder1)
        query = f"title = '{self.title}' and trashed = false"
        count = 0
        for file1 in pydrive_list_item(drive, query):
            self.assertFileInFileList(file1)
            count += 1

        self.assertTrue(count == 11)

    # setUp and tearDown methods.
    # ===========================
    def setUp(self):
        title = test_util.CreateRandomFileName()
        file_list = []
        for x in range(0, 10):
            file1 = self.drive.CreateFile()
            file1["title"] = title
            pydrive_retry(file1.Upload)
            file_list.append(file1)

        self.title = title
        self.file_list = file_list

    def tearDown(self):
        # Deleting uploaded files.
        for file1 in self.file_list:
            pydrive_retry(file1.Delete)

    def assertFileInFileList(self, file_object):
        found = False
        for file1 in self.file_list:
            if file_object["id"] == file1["id"]:
                found = True
        self.assertEqual(found, True)

    def DeleteOldFile(self, file_name):
        try:
            os.remove(file_name)
        except OSError:
            pass


if __name__ == "__main__":
    unittest.main()
